from itertools import chain
rset1 = rql('Any X,A WHERE X ark A, X cwuri ~="None%", X is IN (Agent, ConceptScheme, Concept)')
rset2 = rql('Any X,A WHERE X ark A, X cwuri ~="%s%%", X is IN (Agent, ConceptScheme, Concept)' % config['base-url'])
for entity in chain(rset1.entities(), rset2.entities()):
    assert entity.ark
    entity.cw_set(cwuri=u'ark:/' + entity.ark)
commit()

rql('SET X in_state S WHERE NOT X in_state Z, X is ET, ET default_workflow WF, WF initial_state S')
commit()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref monkeypatch of not-yet-integrated patches"""

from logilab.common.decorators import monkeypatch


# https://www.cubicweb.org/ticket/5352619 ##########################################################
# other part in site_cubicweb.py

from cubicweb import neg_role  # noqa
from cubicweb.web import ProcessFormError  # noqa
from cubicweb.web.views import editcontroller  # noqa

orig_insert_entity = editcontroller.EditController._insert_entity


@monkeypatch(editcontroller.EditController)
def _insert_entity(self, etype, eid, rqlquery):
    if getattr(rqlquery, 'canceled', False):
        return
    return orig_insert_entity(self, etype, eid, rqlquery)


orig_update_entity = editcontroller.EditController._update_entity


@monkeypatch(editcontroller.EditController)
def _update_entity(self, eid, rqlquery):
    if getattr(rqlquery, 'canceled', False):
        return
    return orig_update_entity(self, eid, rqlquery)


@monkeypatch(editcontroller.EditController)
def handle_formfield(self, form, field, rqlquery=None):
    eschema = form.edited_entity.e_schema
    try:
        for field, value in field.process_posted(form):
            if not ((field.role == 'subject' and field.name in eschema.subjrels) or
                    (field.role == 'object' and field.name in eschema.objrels)):
                continue
            rschema = self._cw.vreg.schema.rschema(field.name)
            if rschema.final:
                rqlquery.set_attribute(field.name, value)
            else:
                if form.edited_entity.has_eid():
                    origvalues = set(e.eid for e in form.edited_entity.related(field.name,
                                                                               field.role,
                                                                               entities=True))
                else:
                    origvalues = set()
                if value is None or value == origvalues:
                    continue  # not edited / not modified / to do later
                rdef = eschema.rdef(rschema, field.role)
                if not value and rdef.composite == neg_role(field.role):
                    # deleted composite relation, delete the subject entity
                    self.handle_composite_deletion(form, field, rqlquery)
                if rschema.inlined and rqlquery is not None and field.role == 'subject':
                    self.handle_inlined_relation(form, field, value, origvalues, rqlquery)
                elif form.edited_entity.has_eid():
                    self.handle_relation(form, field, value, origvalues)
                else:
                    form._cw.data['pending_others'].add((form, field))
    except ProcessFormError as exc:
        self.errors.append((field, exc))


@monkeypatch(editcontroller.EditController)
def handle_composite_deletion(self, form, field, rqlquery):
    """handle deletion of the entity when relation to its parent is removed
    """
    rql = 'DELETE %s X WHERE X eid %%(x)s' % form.edited_entity.cw_etype
    self.relations_rql.append((rql, {'x': form.edited_entity.eid}))
    rqlquery.canceled = True



if repo.system_source.dbdriver == 'postgres':
    sql('''
DROP FUNCTION IF EXISTS limit_size (fulltext text, format text, maxsize integer);
CREATE FUNCTION limit_size (fulltext text, format text, maxsize integer) RETURNS text AS $$
DECLARE
    plaintext text;
BEGIN
    IF char_length(fulltext) < maxsize THEN
       RETURN fulltext;
    END IF;
    IF format = 'text/html' OR format = 'text/xhtml' OR format = 'text/xml' THEN
       plaintext := regexp_replace(fulltext, '<[a-zA-Z/][^>]*>', '', 'g');
    ELSE
       plaintext := fulltext;
    END IF;
    IF char_length(plaintext) < maxsize THEN
       RETURN plaintext;
    ELSE
       RETURN substring(plaintext from 1 for maxsize) || '...';
    END IF;
END
$$ LANGUAGE plpgsql;;
''')

for etype in ('Structure', 'History'):
    drop_relation_definition(etype, 'equivalent_concept', 'Concept')
    drop_relation_definition(etype, 'equivalent_concept', 'ExternalUri')
    sync_schema_props_perms(etype)



"""cubicweb-saem_ref unit tests for schema"""

from datetime import date

from cubicweb import ValidationError, Unauthorized, neg_role
from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref import (PERMISSIONS_GRAPHS, mandatory_rdefs,
                            optional_relations)

import testutils


def graph_relations(schema, parent_structure):
    """Given a parent structure of a composite graph (and a schema object),
    return relation information `(rtype, role)` sets where `role` is the role
    of the child in the relation for the following kinds of relations:

    * structural relations,
    * optional relations (cardinality of the child not in '1*'),
    * mandatory relations (cardinality of the child in '1*').
    """
    def concat_sets(sets):
        """Concatenate sets"""
        return reduce(lambda x, y: x | y, sets, set())

    optionals = concat_sets(
        optional_relations(schema, parent_structure).values())
    mandatories = set([
        (rdef.rtype, neg_role(role))
        for rdef, role in mandatory_rdefs(schema, parent_structure)])
    structurals = concat_sets(map(set, parent_structure.values()))
    return structurals, optionals, mandatories


def assertUnauthorized(self, cnx, rql):
    with self.assertRaises(Unauthorized):
        cnx.execute(rql)
        cnx.commit()
    cnx.rollback()


class SchemaConstraintsTC(CubicWebTC):

    def test_on_create_set_end_date_before_start_date(self):
        """ create an entity whose end_date is before start_date.
        ValidationError expected
        """
        with self.admin_access.repo_cnx() as cnx:
            kind = cnx.create_entity('AgentKind', name=u"King")
            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity('Agent', name=u'Arthur', agent_kind=kind,
                                  start_date=date(524, 02, 9), end_date=date(500, 07, 12))
                cnx.commit()
            self.assertIn("must be less than", str(cm.exception))

    def test_on_update_set_end_date_before_start_date(self):
        """ create a valid entity and update it with a new end_date set before the start_date.
            ValidationError expected
        """
        with self.admin_access.repo_cnx() as cnx:
            kind = cnx.create_entity('AgentKind', name=u"King")
            agent = cnx.create_entity('Agent', name=u'Arthur', agent_kind=kind,
                                      start_date=date(454, 02, 9), end_date=date(475, 04, 12))
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                agent.cw_set(end_date=date(442, 07, 12))
                cnx.commit()
            self.assertIn("must be less than", str(cm.exception))

    def test_on_update_set_start_date_after_end_date(self):
        """ create an entity without start_date :
            No constraint on the end_date
            update the entity with a start_date set after the start_date :
            ValidationError expected
        """
        with self.admin_access.repo_cnx() as cnx:
            kind = cnx.create_entity('AgentKind', name=u"Queen")
            agent = cnx.create_entity('Agent', name=u'Guenievre',
                                      agent_kind=kind, end_date=date(476, 02, 9))
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                agent.cw_set(start_date=date(527, 04, 12))
                cnx.commit()
            self.assertIn("must be less than", str(cm.exception))

    def test_published_constraint_on_contact_point(self):
        """ create two agents: one published P and one not published N.
            create one more agent and check that interface will only show P that can become its
            contact point and archival agent
        """
        with self.admin_access.repo_cnx() as cnx:
            peter = testutils.agent(cnx, u'Peter', archival_roles=[u'archival'])
            testutils.agent(cnx, u'Norton', archival_roles=[u'archival'])
            cnx.commit()
            peter.cw_adapt_to('IWorkflowable').fire_transition('publish')
            agent = testutils.agent(cnx, u'Alice')
            cnx.commit()
            for rtype in ('archival_agent', 'contact_point'):
                rset = agent.unrelated(rtype, 'Agent')
                self.assertEqual(rset.one().eid, peter.eid)

    def test_not_user_constraint_on_agent_related_list(self):
        """ create two agents: one linked to a user U and one not linked to any user N.
            create one more agent and check that interface will only show N that can become its
            contact point and archival agent
        """
        with self.admin_access.repo_cnx() as cnx:
            # Create a CWUser: an associated agent will be automatically created and linked to it
            users = cnx.find('CWGroup', name=u'users').one()
            jdoe_user = cnx.create_entity('CWUser', login=u'jdoe', upassword=u'jdoe',
                                          in_group=users)
            cnx.commit()
            # Give archival role to this agent
            jdoe = cnx.find('Agent', agent_user=jdoe_user).one()
            archival_role = cnx.find('ArchivalRole', name=u'archival').one()
            jdoe.cw_set(archival_role=archival_role)
            cnx.commit()
            # Create another archival agent, not related to a user
            norton = testutils.agent(cnx, u'Norton', archival_roles=[u'archival'])
            cnx.commit()
            # Publish both agents (because of the published constraint)
            jdoe.cw_adapt_to('IWorkflowable').fire_transition('publish')
            norton.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            # Now create a third agent and check the unrelated list
            agent = testutils.agent(cnx, u'Alice')
            cnx.commit()
            for rtype in ('archival_agent', 'contact_point'):
                rset = agent.unrelated(rtype, 'Agent')
                self.assertEqual(rset.one().eid, norton.eid)


class AgentTC(CubicWebTC):
    assertUnauthorized = assertUnauthorized

    def test_fti(self):
        with self.admin_access.repo_cnx() as cnx:
            kind = cnx.create_entity('AgentKind', name=u"Queen")
            agent = cnx.create_entity('Agent', name=u'Guenievre',
                                      agent_kind=kind, end_date=date(476, 02, 9))
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', name=u'place', place_address=address, place_agent=agent)
            cnx.create_entity('AgentFunction', name=u'function', function_agent=agent)
            cnx.create_entity('LegalStatus', term=u'legal status', legal_status_agent=agent)
            cnx.commit()
            for search in (u'guenievre', u'europe', u'place', u'function', u'legal status'):
                self.set_description(search)
                # TODO: use subTest
                self.assertEqual(cnx.execute('Agent X WHERE X has_text %(search)s',
                                             {'search': search}).one().eid, agent.eid)

    def test_graph_structure(self):
        graph = PERMISSIONS_GRAPHS['Agent'](self.schema)
        expected = {
            'AgentFunction': {('function_agent', 'subject'): set(['Agent'])},
            'AgentPlace': {('place_agent', 'subject'): set(['Agent'])},
            'Citation': {('has_citation', 'object'): set(['Mandate', 'Occupation'])},
            'EACResourceRelation': {('resource_relation_agent', 'subject'): set(['Agent'])},
            'EACSource': {('source_agent', 'subject'): set(['Agent'])},
            'History': {('history_agent', 'subject'): set(['Agent'])},
            'LegalStatus': {('legal_status_agent', 'subject'): set(['Agent'])},
            'Mandate': {('mandate_agent', 'subject'): set(['Agent'])},
            'Occupation': {('occupation_agent', 'subject'): set(['Agent'])},
            'PhoneNumber': {('phone_number', 'object'): set(['Agent'])},
            'PostalAddress': {('place_address', 'object'): set(['AgentPlace'])},
            'Structure': {('structure_agent', 'subject'): set(['Agent'])},
        }
        struct = dict(
            (k, dict((rel, set(targets)) for rel, targets in v.items()))
             for k, v in graph.parent_structure('Agent').items())
        self.assertEqual(struct, expected)

    def test_optional_relations(self):
        graph = PERMISSIONS_GRAPHS['Agent'](self.schema)
        structure = graph.parent_structure('Agent')
        opts = optional_relations(self.schema, structure)
        expected = {'Citation': set([('has_citation', 'object')])}
        self.assertEqual(opts, expected)

    def test_relations_consistency(self):
        graph = PERMISSIONS_GRAPHS['Agent'](self.schema)
        structure = graph.parent_structure('Agent')
        structurals, optionals, mandatories = graph_relations(
            self.schema, structure)
        self.assertEqual(structurals - optionals, mandatories)

    def test_security_base(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', 'guests'))
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            cnx.create_entity(
                'Agent', name=u'bob',
                reverse_function_agent=cnx.create_entity('AgentFunction', name=u'grouillot'),
                agent_kind=cnx.find('AgentKind', name=u'person').one())
            cnx.commit()
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'DELETE U in_group G WHERE U login "toto", G name "users"')
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            self.assertUnauthorized(
                cnx, 'SET X name "bobby" WHERE X is Agent, X name "bob"')
            self.assertUnauthorized(
                cnx, 'SET X agent_kind K WHERE X name "bob", K name "authority"')
            self.assertUnauthorized(
                cnx, 'SET F name "director" WHERE X name "bob", F function_agent X')

    def test_security_published(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', ))
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            agent = cnx.create_entity(
                'Agent', name=u'bob',
                agent_kind=cnx.find('AgentKind', name=u'person').one())
            cnx.commit()
            iwf = agent.cw_adapt_to('IWorkflowable')
            iwf.fire_transition('publish')
            cnx.commit()
            # we can still modify a published agent
            cnx.execute('SET X name "bobby" WHERE X is Agent, X name "bob"')
            cnx.execute('SET X agent_kind K WHERE X name "bob", K name "authority"')
            cnx.execute('SET F name "director" WHERE X name "bob", F function_agent X')
            cnx.commit()


class ConceptSchemeTC(CubicWebTC):

    def test_graph_structure(self):
        graph = PERMISSIONS_GRAPHS['ConceptScheme'](self.schema)
        expected = {
            'Concept': {('in_scheme', 'subject'): ['ConceptScheme']},
            'Label': {('label_of', 'subject'): ['Concept']},
        }
        self.assertEqual(graph.parent_structure('ConceptScheme'),
                         expected)

    def test_optional_relations(self):
        graph = PERMISSIONS_GRAPHS['ConceptScheme'](self.schema)
        opts = optional_relations(self.schema,
                                  graph.parent_structure('ConceptScheme'))
        expected = {}
        self.assertEqual(opts, expected)

    def test_relations_consistency(self):
        graph = PERMISSIONS_GRAPHS['ConceptScheme'](self.schema)
        structure = graph.parent_structure('ConceptScheme')
        structurals, optionals, mandatories = graph_relations(
            self.schema, structure)
        self.assertEqual(structurals - optionals, mandatories)


class SEDAProfileTC(CubicWebTC):

    def test_fti(self):
        # "Reverse" text to be searched in order not to be troubled by other
        # entities that may live in the DB (e.g. Concepts) with similar text.
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            archive = profile.archives[0]
            archive.name.cw_set(value=u'profile name'[::-1])
            ac = archive.access_restriction_code
            ac.cw_set(user_annotation=u'some annotation'[::-1])
            seda_document_type_code_value = cnx.execute(
                'Concept X WHERE X preferred_label L, L label "CDO"').one()
            seda_document_type_code = cnx.create_entity(
                'SEDADocumentTypeCode',
                seda_document_type_code_value=seda_document_type_code_value)
            cnx.create_entity('ProfileDocument',
                              seda_description=cnx.create_entity(
                                  'SEDADescription',
                                  value=u'document description'[::-1]),
                              seda_document_type_code=seda_document_type_code,
                              seda_parent=archive)
            cnx.commit()
            for search in ('name', 'annotation', 'description'):
                self.set_description(search)
                rset = cnx.execute('Any X WHERE X has_text %(search)s',
                                   {'search': search[::-1]})
                # TODO: use subTest
                self.assertEqual([r for r, in rset.rows], [profile.eid])

    def test_graph_structure(self):
        graph = PERMISSIONS_GRAPHS['ProfileArchiveObject'](self.schema)
        expected = {
            'ProfileArchiveObject': {
                ('seda_parent', 'subject'): [
                    'SEDAProfile', 'ProfileArchiveObject']},
            'ProfileDocument': {
                ('seda_parent', 'subject'): ['ProfileArchiveObject']},
            'SEDAAccessRestrictionCode': {
                ('seda_access_restriction_code', 'object'): [
                    'ProfileArchiveObject']},
            'SEDAAppraisalRule': {
                ('seda_appraisal_rule', 'object'): ['ProfileArchiveObject']},
            'SEDAAppraisalRuleCode': {
                ('seda_appraisal_rule_code', 'object'): ['SEDAAppraisalRule']},
            'SEDAAppraisalRuleDuration': {
                ('seda_appraisal_rule_duration', 'object'): [
                    'SEDAAppraisalRule']},
            'SEDACharacterSetCode': {
                ('seda_character_set_code', 'object'): ['ProfileDocument']},
            'SEDAContentDescription': {
                ('seda_content_description', 'object'): [
                    'ProfileArchiveObject']},
            'SEDADate': {
                ('seda_latest_date', 'object'): ['SEDAContentDescription'],
                ('seda_oldest_date', 'object'): ['SEDAContentDescription']},
            'SEDADescription': {
                ('seda_description', 'object'): ['SEDAContentDescription',
                                                 'ProfileDocument']},
            'SEDADescriptionLevel': {
                ('seda_description_level', 'object'): [
                    'SEDAContentDescription']},
            'SEDADocumentTypeCode': {
                ('seda_document_type_code', 'object'): ['ProfileDocument']},
            'SEDAFileTypeCode': {
                ('seda_file_type_code', 'object'): ['ProfileDocument']},
            'SEDAKeyword': {
                ('seda_keyword_of', 'subject'): ['SEDAContentDescription']},
            'SEDAName': {
                ('seda_name', 'object'): ['ProfileArchiveObject']},
        }
        self.assertEqual(graph.parent_structure('SEDAProfile'),
                         expected)

    def test_optional_relations(self):
        graph = PERMISSIONS_GRAPHS['ProfileArchiveObject'](self.schema)
        opts = optional_relations(self.schema,
                                  graph.parent_structure('SEDAProfile'))
        expected = {
            'ProfileArchiveObject': set([('seda_parent', 'subject')]),
            'ProfileDocument': set([('seda_parent', 'subject')]),
            'SEDADate': set([('seda_oldest_date', 'object'),
                             ('seda_latest_date', 'object')]),
        }
        self.assertEqual(opts, expected)

    def test_relations_consistency(self):
        graph = PERMISSIONS_GRAPHS['ProfileArchiveObject'](self.schema)
        structure = graph.parent_structure('SEDAProfile')
        structurals, optionals, mandatories = graph_relations(
            self.schema, structure)
        self.assertEqual(structurals - optionals, mandatories)

    def _test_relative_permissions_through(self, permission):
        actual = [e.expression
                  for e in self.schema['ProfileArchiveObject'].permissions[permission]]
        expected = [
            u'NOT EXISTS(X seda_parent A), U in_group G, G name IN("managers", "users")',
            u'X seda_parent A, U has_update_permission A',
        ]
        self.assertEqual(actual, expected)
        actual = [e.expression
                  for e in self.schema['SEDADate'].permissions[permission]]
        expected = [
            'A seda_oldest_date X, U has_update_permission A',
            'A seda_latest_date X, U has_update_permission A',
        ]
        self.assertEqual(actual, expected)

    def test_relative_update_permissions_through(self):
        self._test_relative_permissions_through('update')

    def test_relative_delete_permissions_through(self):
        self._test_relative_permissions_through('delete')

    def test_relative_permissions_of(self):
        rdef = self.schema['seda_parent'].rdef('ProfileDocument', 'ProfileArchiveObject')
        rqlexpr, = rdef.permissions['add']
        self.assertEqual(rqlexpr.expression, 'U has_update_permission O')
        rdef = self.schema['seda_oldest_date'].rdef('SEDAContentDescription', 'SEDADate')
        rqlexpr, = rdef.permissions['delete']
        self.assertEqual(rqlexpr.expression, 'U has_update_permission S')

    def test_dates_constraints(self):
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            archive = profile.archives[0]
            desc = archive.content_description
            date = cnx.create_entity('SEDADate',
                                     reverse_seda_oldest_date=desc)
            cnx.commit()

            def assert_error(cm, date_type):
                self.assertEqual(cm.exception.errors,
                                 {'seda_%s_date-object' % date_type:
                                  u'SEDA date cannot be used as oldest and latest'})

            datetype = 'latest'
            self.set_description(datetype)
            with self.assertRaises(ValidationError) as cm:
                desc.cw_set(seda_latest_date=date)
                cnx.commit()
            cnx.rollback()
            assert_error(cm, datetype)
            desc.cw_set(seda_oldest_date=None)
            desc.cw_set(seda_latest_date=date)
            cnx.commit()
            datetype = 'oldest'
            self.set_description(datetype)
            with self.assertRaises(ValidationError) as cm:
                desc.cw_set(seda_oldest_date=date)
                cnx.commit()
            cnx.rollback()
            assert_error(cm, datetype)


class SEDAKeywordTC(CubicWebTC):

    def test_scheme_value_constraint(self):
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            cnx.commit()
            cd = profile.archives[0].content_description
            concept = cnx.execute('Concept X WHERE X preferred_label L, L label "file"').one()
            scheme = cnx.execute('ConceptScheme X LIMIT 1 WHERE NOT X eid %(x)s',
                                 {'x': concept.scheme.eid}).one()
            with self.assertRaises(ValidationError) as cm:
                kw = cnx.create_entity('SEDAKeyword', seda_keyword_of=cd)
                kw.cw_set(seda_keyword_value=concept)
                kw.cw_set(seda_keyword_scheme=scheme)
                cnx.commit()
            self.assertEqual(cm.exception.errors,
                             {'seda_keyword_value-subject':
                              u"concept doesn't belong to the scheme associated to the keyword"})
            cnx.rollback()
            kw = cnx.create_entity('SEDAKeyword', seda_keyword_of=cd)
            kw.cw_set(seda_keyword_value=concept,
                      seda_keyword_scheme=concept.scheme)
            cnx.commit()


class WorkflowPermissionsTC(CubicWebTC):
    """Test case for permissions on entity type with "publication" workflow.
    """
    assertUnauthorized = assertUnauthorized

    def test_update_root_badgroup(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', 'guests'))
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            testutils.setup_seda_profile(cnx, title=u'pp')
            cnx.commit()
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'DELETE U in_group G WHERE U login "toto", G name "users"')
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            self.assertUnauthorized(cnx, 'SET X title "qq" WHERE X is SEDAProfile, X title "pp"')

    def test_update_child(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', ))
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            agent = cnx.create_entity(
                'Agent', name=u'bob',
                agent_kind=cnx.find('AgentKind', name=u'person').one())
            cnx.create_entity('AgentFunction', name=u'boss',
                              function_agent=agent)
            cnx.commit()
            # Draft -> OK.
            cnx.execute('SET X name "grouyo" WHERE X function_agent A')
            cnx.commit()
            cnx.create_entity('PhoneNumber', number=u'1234',
                              reverse_phone_number=agent)
            cnx.commit()
            iwf = agent.cw_adapt_to('IWorkflowable')
            iwf.fire_transition('publish')
            cnx.commit()
            # Published -> still OK for agent.
            cnx.execute('SET X name "big boss" WHERE X function_agent A')
            cnx.commit()
            cnx.create_entity('PhoneNumber', number=u'5678',
                              reverse_phone_number=agent)
            cnx.commit()

    def test_sedaprofile_permissions(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx, title=u'pp')
            name = cnx.create_entity('SEDAName')
            profile_archive_obj = cnx.create_entity(
                'ProfileArchiveObject', seda_name=name, seda_parent=profile)
            cnx.commit()
            # Profile in draft, modifications are allowed.
            profile.cw_set(title=u'ugh')
            cnx.commit()
            name.cw_set(value=u'argh')
            cnx.commit()
            seda_date = cnx.create_entity('SEDADate')
            profile_archive_obj.content_description.cw_set(seda_oldest_date=seda_date)
            cnx.commit()
            # Profile published, no modification allowed.
            profile.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            with self.assertRaises(Unauthorized):
                profile.cw_set(title=u'gloups')
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                name.cw_set(value=u'zorglub')
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                seda_date.cw_set(value=date.today())
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                profile_archive_obj.content_description.cw_set(
                    seda_latest_date=cnx.create_entity('SEDADate'))
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                seda_date.cw_delete()
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                profile_archive_obj.cw_delete()
                cnx.commit()
            cnx.rollback()

    def test_sedaprofile_no_parent_permissions(self):
        with self.admin_access.repo_cnx() as cnx:
            name = cnx.create_entity('SEDAName')
            profile_archive_obj = cnx.create_entity(
                'ProfileArchiveObject', seda_name=name)
            cnx.commit()
            # document unit has no parent, modifications are allowed.
            name.cw_set(value=u'argh')
            cnx.commit()
        with self.new_access('anon').client_cnx() as cnx:
            name = cnx.entity_from_eid(name.eid)
            profile_archive_obj = cnx.entity_from_eid(profile_archive_obj.eid)
            with self.assertRaises(Unauthorized):
                name.cw_set(value=u'zorglub')
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                profile_archive_obj.cw_set(user_annotation=u'zorglub')
                cnx.commit()
            cnx.rollback()
            ar = cnx.create_entity('SEDAAccessRestrictionCode')
            profile_archive_obj.cw_set(seda_access_restriction_code=ar)
            with self.assertRaises(Unauthorized):
                cnx.commit()
            cnx.rollback()
        with self.admin_access.repo_cnx() as cnx:
            profile_archive_obj = cnx.entity_from_eid(profile_archive_obj.eid)
            profile_archive_obj.cw_delete()
            cnx.commit()

    def test_conceptscheme_permissions(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.execute('Any X LIMIT 1 WHERE X is ConceptScheme').one()
            # in draft, modifications are allowed.
            concept = scheme.add_concept(u'blah')
            cnx.commit()
            # published, can't modify existing content.
            scheme.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            with self.assertRaises(Unauthorized):
                scheme.cw_set(description=u'plop')
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                concept.preferred_label[0].cw_set(label=u'plop')
                cnx.commit()
            cnx.rollback()
            # though addition of new concepts / labels is fine
            new_concept = scheme.add_concept(u'plop')
            cnx.commit()
            new_label = cnx.create_entity('Label', label=u'arhg', label_of=concept)
            cnx.commit()
            # while deletion is fine for label but not for concept nor scheme
            new_label.cw_delete()
            cnx.commit()
            with self.assertRaises(Unauthorized):
                scheme.cw_delete()
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                new_concept.cw_delete()
                cnx.commit()
            cnx.rollback()


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views/forms/actions/components for web ui"""

from functools import wraps

from logilab.common.decorators import monkeypatch
from logilab.mtconverter import xml_escape

from cubicweb import tags
from cubicweb.uilib import js, domid
from cubicweb.utils import JSString
from cubicweb.view import EntityView
from cubicweb.predicates import has_permission, is_instance, multi_lines_rset, match_kwargs
from cubicweb.web import component
from cubicweb.web.views import uicfg, actions, urlrewrite, tabs, primary, debug

from cubes.squareui.views.basetemplates import basetemplates

from cubes.saem_ref import cwuri_url


_ = unicode


def external_link(text, href):
    """Return an HTML link with an icon indicating that the URL is external.
    """
    icon = tags.span(klass="glyphicon glyphicon-share-alt")
    return tags.a(u' '.join([text, icon]), href=href, escapecontent=False)


def add_etype_link(req, etype, text=u'', klass='icon-plus-circled pull-right',
                   **urlparams):
    """Return an HTML link to add an entity of type 'etype'."""
    vreg = req.vreg
    eschema = vreg.schema.eschema(etype)
    if eschema.has_perm(req, 'add'):
        url = vreg['etypes'].etype_class(etype).cw_create_url(req, **urlparams)
        return tags.a(text, href=url, klass=klass,
                      title=req.__('New %s' % etype))
    return u''


def import_etype_link(req, etype, url):
    """Return an HTML link to the view that may be used to import an entity of type `etype`.
    """
    eschema = req.vreg.schema.eschema(etype)
    if eschema.has_perm(req, 'add'):
        return tags.a(u'', href=url, klass='icon-upload pull-right',
                      title=req.__('Import %s' % etype))
    return u''


def index_dropdown_button(text, links):
    """Return an HTML button with `text` and dropdown content from `links`, for use in an index box
    title.
    """
    links = [l for l in links if l]  # there may be empty links for not allowed actions
    if not links:
        return u''  # no need for the dropdown
    data = []
    w = data.append
    w(u'<div class="btn-group pull-right clearfix">')
    w(tags.a('', href='', klass='icon-plus-circled pull-right dropdown-toggle',
             title=text, **{'data-toggle': 'dropdown', 'aria-expanded': 'false'}))
    w(u'<ul class="dropdown-menu" role="menu">')
    for link in links:
        w(u'<li>{0}</li>'.format(link))
    w(u'</ul>')
    w(u'</div>')
    return u''.join(data)


def editlinks(icon_info):
    """Decorator for `entity_call` to add "edit" links."""
    def decorator(entity_call):
        @wraps(entity_call)
        def wrapper(self, entity, **kwargs):
            editurlparams = {
                '__redirectpath': kwargs.pop('__redirectpath',
                                             entity.rest_path()),
            }
            if 'tabid' in kwargs:
                editurlparams['__redirectparams'] = 'tab=' + kwargs.pop('tabid')
            html_icons = entity.view('saem_ref.edit-link',
                                     icon_info=icon_info,
                                     editurlparams=editurlparams)
            if html_icons:
                self.w(tags.div(html_icons, klass='pull-right'))
            return entity_call(self, entity, **kwargs)
        return wrapper
    return decorator


class ExternalLinkView(EntityView):
    """Render an entity as an external link according to `rtype` keyword
    argument or automatically if it is an ExternalUri.
    """

    __regid__ = 'saem.external_link'
    __select__ = EntityView.__select__ & (
        is_instance('ExternalUri') | match_kwargs('rtype'))

    def entity_call(self, entity, rtype=None):
        if rtype is None:
            rtype = 'uri'  # ExternalUri
        href = text = getattr(entity, rtype)
        self.w(external_link(text, href))


def configure_relation_widget(req, div, search_url, title, multiple, validate):
    """Build a javascript link to invoke a relation widget

    Widget will be linked to div `div`, with a title `title`. It will display selectable entities
    matching url `search_url`. bool `multiple` indicates whether several entities can be selected or
    just one, `validate` identifies the javascript callback that must be used to validate the
    selection.
    """
    req.add_js(('jquery.ui.js',
                'cubicweb.ajax.js',
                'cubicweb.widgets.js',
                'cubicweb.facets.js',
                'cubes.relationwidget.js',
                'cubes.saem_ref.js'))
    return 'javascript: %s' % js.saem.relateWidget(div, search_url, title, multiple,
                                                   JSString(validate))


class AddEntityComponent(component.CtxComponent):
    """Component with 'add' link to be displayed in 'same etype' views usually 'SameETypeListView'.
    """
    __regid__ = 'saem_ref.add_entity'
    __select__ = (component.CtxComponent.__select__ & multi_lines_rset() & has_permission('add') &
                  is_instance('Agent', 'ConceptScheme', 'SEDAProfile'))
    context = 'navtop'

    def render_body(self, w):
        etype = self.cw_rset.description[0][0]
        w(add_etype_link(self._cw, etype))


class ImportEntityComponent(component.CtxComponent):
    """Component with 'import' link to be displayed in 'same etype' views usually
    'SameETypeListView'.

    Concret class should fill the `import_vid` class attribute and add a proper `is_instance`
    selector.
    """
    __abstract__ = True
    __regid__ = 'saem_ref.import_entity'
    __select__ = component.CtxComponent.__select__ & multi_lines_rset() & has_permission('add')
    import_url = None  # URL of the view that may be used to import data
    context = 'navtop'

    def render_body(self, w):
        etype = self.cw_rset.description[0][0]
        w(import_etype_link(self._cw, etype, self.import_url))


# Bootstrap configuration.
basetemplates.TheMainTemplate.twbs_container_cls = 'container'
basetemplates.HTMLPageHeader.css = {
    'navbar-extra': 'navbar-default',
    'breadcrumbs': 'cw-breadcrumb',
    'container-cls': basetemplates.TheMainTemplate.twbs_container_cls,
    'header-left': '',
    'header-right': 'navbar-right',
}


# Wrap breadcrumbs items within a "container" div.
@monkeypatch(basetemplates.HTMLPageHeader)
def breadcrumbs(self, view):
    components = self.get_components(view, context='header-center')
    if components:
        self.w(u'<nav role="navigation" class="%s">' %
               self.css.get('breadcrumbs', 'breadcrumbs-defaul'))
        for comp in components:
            self.w(u'<div class="container">')
            comp.render(w=self.w)
            self.w(u'</div>')
        self.w(u'</nav>')


for etype in ('Agent', 'ConceptScheme', 'SEDAProfile'):
    uicfg.autoform_section.tag_subject_of(
        (etype, 'custom_workflow', '*'), 'main', 'hidden')


class EditLinkView(EntityView):
    """Render entity as a link to the edition view"""
    __regid__ = 'saem_ref.edit-link'

    def entity_call(self, entity, icon_info=True, editurlparams=None):
        editurlparams = editurlparams or {}
        links = []
        if icon_info:
            links.append(('icon-info', self._cw._('view'), entity.absolute_url()))
        if entity.cw_has_perm('update'):
            links.append(('icon-pencil', self._cw._('edit'),
                          entity.absolute_url(vid='edition', **editurlparams)))
        if entity.cw_has_perm('delete'):
            links.append(('icon-trash', self._cw._('delete'),
                          entity.absolute_url(vid='deleteconf', **editurlparams)))
        for csscls, label, href in links:
            self.w(u'<span title="{label}">{link}</span>'.format(
                label=label, link=tags.a(klass=csscls, href=href)))


urlrewrite.SchemaBasedRewriter.rules.append(
    (urlrewrite.rgx('/ark:/(.+)'),
     urlrewrite.build_rset(rql=r'Any X WHERE X ark %(text)s', rgxgroups=[('text', 1)]))
)


class SAEMRefRewriter(urlrewrite.SimpleReqRewriter):
    """URL rewriter for SAEM-Ref"""
    rules = [
        ('/oai', dict(vid='oai')),
        ('/ark', dict(vid='saem_ref.ws.ark')),
        ('/sedalib', dict(rql='Any X WHERE X is in (ProfileArchiveObject, ProfileDocument), '
                              'NOT X seda_parent P',
                          vid='saem_ref.sedalib')),
    ]


@monkeypatch(tabs.TabsMixin)
def active_tab(self, default):
    return self._cw.form.get('tab', domid(default))


class URLAttributeView(primary.URLAttributeView):
    """Overriden from cubicweb to handle case where ark is directly used in cwuri, and so absolute
    url should be generated from it.
    """
    def entity_call(self, entity, rtype, **kwargs):
        if rtype == 'cwuri':
            url = xml_escape(cwuri_url(entity))
        else:
            url = entity.printable_value(rtype)
        if url:
            self.w(u'<a href="%s">%s</a>' % (url, url))


actions.SiteConfigurationAction.order = 100
debug.SiteInfoAction.order = 101


def registration_callback(vreg):
    from cubicweb.web.views import tableview, basecomponents, undohistory
    vreg.register_all(globals().values(), __name__, (URLAttributeView,))
    vreg.register_and_replace(URLAttributeView, primary.URLAttributeView)
    vreg.unregister(tableview.TableView)
    vreg.unregister(undohistory.UndoHistoryView)
    vreg.unregister(basecomponents.ApplicationName)
    # disable 'add a etype' action replaced by appropriate ctx components for
    # scheme and agent, and providing poor user experience anyway (hidden
    # in the actions box)
    vreg.unregister(actions.AddNewAction)



"""cubicweb-saem_ref common test tools"""

import unittest
from doctest import Example

from lxml import etree
from lxml.doctestcompare import LXMLOutputChecker

from logilab.common import tempattr


class Mock(object):
    _counter = 0  # store counter on the class so we don't generate the same identifier in case of
                  # several instantiation of the mock in the same test
    events = []

    @classmethod
    def mock_applied(cls, obj, attrname):
        """Return a context manager within which the mock is replacing `attrname` on the given
        object
        """
        cls.reset_counter()
        cls.reset_events()
        return tempattr(obj, attrname, cls)

    @classmethod
    def reset_counter(cls):
        cls._counter = 0

    @classmethod
    def reset_events(cls):
        cls.events = []

    @classmethod
    def _create_ref(cls):
        cls._counter += 1
        return cls._counter

    def __init__(self, *args, **kwargs):
        self._record('__init__', *args, **kwargs)

    def _record(self, __event, *args, **kwargs):
        ev_descr = [__event]
        ev_descr += args
        if kwargs:
            ev_descr.append(kwargs)
        self.events.append(tuple(ev_descr))


def agent(cnx, name, kind=u'person', archival_roles=(), **kwargs):
    """Return an Agent with specified kind, name and archival roles."""
    kind_eid = cnx.find('AgentKind', name=kind)[0][0]
    if archival_roles:
        if len(archival_roles) > 1:
            rset = cnx.execute('Any X WHERE X is ArchivalRole, X name IN (%s)' %
                               ','.join('"%s"' % r for r in archival_roles))
        else:
            rset = cnx.execute('Any X WHERE X is ArchivalRole, X name "%s"' %
                               archival_roles[0])
        roles_eid = [x for x, in rset.rows]
    else:
        roles_eid = ()
    return cnx.create_entity('Agent', name=name, agent_kind=kind_eid,
                             archival_role=roles_eid, **kwargs)


def setup_seda_agents(cnx):
    """Return two agents, one with the 'deposit' archival role, another with
    the 'archival' archival role.
    """
    alice = agent(cnx, u'alice')
    city = agent(cnx, u'New York city', kind=u'authority',
                 archival_roles=[u'archival'], contact_point=alice)
    bob = agent(cnx, u'bob', archival_roles=['deposit'], archival_agent=city)
    return bob, city


def setup_seda_profile(cnx, complete_archive=True, **kwargs):
    """Return a minimal SEDA profile with agents and a profile archive (with or without content
    description and access restriction code according to the `complete_archive` flag).
    """
    transferring_agent, archival_agent = setup_seda_agents(cnx)
    create_entity = cnx.create_entity
    profile = create_entity('SEDAProfile', seda_transferring_agent=transferring_agent, **kwargs)
    kwargs = {}
    if complete_archive:
        kwargs['seda_content_description'] = create_entity(
            'SEDAContentDescription',
            seda_description_level=create_entity('SEDADescriptionLevel'))
        kwargs['seda_access_restriction_code'] = create_entity('SEDAAccessRestrictionCode')
    create_entity('ProfileArchiveObject', seda_name=create_entity('SEDAName'),
                  seda_parent=profile, **kwargs)
    return profile


def publishable_profile(cnx, **kwargs):
    """Return a minimal SEDA profile that may be published."""
    create_entity = cnx.create_entity
    profile = create_entity('SEDAProfile', **kwargs)
    create_entity('ProfileArchiveObject',
                  seda_parent=profile,
                  seda_name=create_entity('SEDAName'),
                  seda_content_description=create_entity(
                      'SEDAContentDescription',
                      seda_description_level=create_entity('SEDADescriptionLevel')),
                  seda_access_restriction_code=create_entity('SEDAAccessRestrictionCode'))
    return profile


def concept(cnx, label):
    """Return concept entity with the given preferred label (expected to be unique)."""
    return cnx.execute('Concept X WHERE X preferred_label L, L label %(label)s',
                       {'label': label}).one()


#
# XML helpers
#

def xmlpp(string):
    """Parse and pretty-print XML data from `string`."""
    print etree.tostring(etree.fromstring(string), pretty_print=True)


class XmlTestMixin(unittest.TestCase):
    """Mixin class provinding additional assertion methods for checking XML data."""

    def assertXmlEqual(self, actual, expected):
        """Check that both XML strings represent the same XML tree."""
        checker = LXMLOutputChecker()
        if not checker.check_output(expected, actual, 0):
            message = checker.output_difference(Example("", expected), actual, 0)
            self.fail(message)

    def assertXmlValid(self, xml_data, xsd_filename, debug=False):
        """Validate an XML file (.xml) according to an XML schema (.xsd)."""
        with open(xsd_filename) as xsd:
            xmlschema = etree.XMLSchema(etree.parse(xsd))
        # Pretty-print xml_data to get meaningfull line information.
        xml_data = etree.tostring(etree.fromstring(xml_data), pretty_print=True)
        xml_data = etree.fromstring(xml_data)
        if debug and not xmlschema.validate(xml_data):
            xmlpp(xml_data)
        xmlschema.assertValid(xml_data)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref SKOS entity's classes"""

from cubicweb.predicates import is_instance, score_entity

from cubes.skos.entities import ConceptSchemeRDFPrimaryAdapter
from cubes.saem_ref.entities import oai


# prefetch 'ark' attribute of concept for RDF export
ConceptSchemeRDFPrimaryAdapter.concept_attributes.append('ark')


class ConceptSchemeOAIPMHRecordAdapter(oai.OAIPMHActiveRecordAdapter):
    """OAI-PMH adapter for ConceptScheme entity type."""
    __select__ = oai.OAIPMHActiveRecordAdapter.__select__ & is_instance('ConceptScheme')
    metadata_view = 'list.rdf'

    @classmethod
    def set_definition(cls):
        return oai.PublicETypeOAISetSpec(
            'ConceptScheme', cls.identifier_attribute)


class ConceptInPublicSchemeETypeOAISetSpec(oai.ETypeOAISetSpec):
    """OAI-PMH set specifier matching on "Concept" entity type and fetching
    concepts related to scheme not in "draft" state.
    """

    def __init__(self):
        super(ConceptInPublicSchemeETypeOAISetSpec, self).__init__(
            'Concept', 'ark')

    def setspec_restrictions(self, value=None):
        return 'X in_scheme Y, Y in_state ST, NOT ST name "draft"', {}


class ConceptInPublicSchemeRelatedEntityOAISetSpec(oai.RelatedEntityOAISetSpec):
    """OAI-PMH second-level set specifier to match "Concept" entity type
    related to scheme not in "draft" state.
    """

    def __init__(self):
        super(ConceptInPublicSchemeRelatedEntityOAISetSpec, self).__init__(
            'in_scheme', 'ConceptScheme', 'ark',
            description=u'A concept in scheme identified by {0}')

    def setspecs(self, cnx):
        rset = cnx.execute('Any A WHERE X is ConceptScheme, X ark A,'
                           '            X in_state ST, NOT ST name "draft"')
        for value, in rset.rows:
            yield value, self.description.format(value)


def not_in_draft_scheme(concept):
    """Return True if `concept` is in a ConceptScheme not in "draft" state.
    """
    return concept.in_scheme[0].in_state[0].name != 'draft'


class ConceptOAIPMHRecordAdapter(oai.ArkOAIPMHRecordAdapter):
    """OAI-PMH adapter for ConceptScheme entity type."""
    __select__ = (oai.ArkOAIPMHRecordAdapter.__select__ & is_instance('Concept') &
                  score_entity(not_in_draft_scheme))
    metadata_view = 'primary.rdf'

    @classmethod
    def set_definition(cls):
        specifier = ConceptInPublicSchemeETypeOAISetSpec()
        specifier['in_scheme'] = ConceptInPublicSchemeRelatedEntityOAISetSpec()
        return specifier



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Auto-completion in search widgets."""

from collections import defaultdict

from logilab.mtconverter import xml_escape

from cubicweb import Unauthorized
from cubicweb.uilib import domid
from cubicweb.utils import json_dumps, make_uid
from cubicweb.view import StartupView
from cubicweb.web import facet as facetbase
from cubicweb.web.views.boxes import SearchBox
from cubicweb.web.views.facets import HasTextFacet

__docformat__ = "restructuredtext en"

# XXX The words selection should be done in RQL ::
#     Any X HAVING W=WORDS(%(q)s)
#     or
#     Any W, SIMILARITY(W) HAVING W=WORDS(%(q)s)
#
# However, with WORDS class declared as above::
#
#     class WORDS(FunctionDescr):
#         supported_backends = ('postgres',)
#         rtype = 'String'
#
# the resulting sql for Any X HAVING W=WORDS('toto') states :
#
#       SELECT _W.eid
#       FROM entities AS _W
#       WHERE _W.eid=WORDS(%(140390076097112)s) {'140390076097112': u'toto'}
#
# Some more modifications are needed in `server/sources/rql2sql.py`,
# which we dont't want to do for instance.


def get_results_from_query(cnx, param):
    return cnx.system_sql('''SELECT word, etype, similarity(word, %(q)s) AS sml
FROM words
WHERE similarity(word, %(q)s) > 0.4
AND length(word) > length(%(q)s)-1
ORDER BY sml DESC, word, etype
LIMIT 20;''', param).fetchall()


class AutocompleteSearchView(StartupView):
    """The principal view for autocomplete search
    """
    __regid__ = 'search-autocomplete'
    templatable = False
    binary = True

    def rqlexec(self, rql, args=None):
        """Utility method to execute some rql queries, and simply returning an
        empty list if :exc:`Unauthorized` is raised.
        """
        try:
            return self._cw.execute(rql, args)
        except Unauthorized:
            return []

    def call(self):
        rset = get_results_from_query(self._cw.cnx, {'q': self._cw.form.get('term')})
        if not rset:
            self.w(json_dumps([]))
        results = []
        rql = self._cw.form['facetrql']
        all_words = ' '.join(self._cw.form['all'].split()[:-1])
        if rql:
            res = self._build_results_for_facet(rset, rql, all_words)
        else:
            res = self._build_results_for_search(rset, all_words)
        for word, labels in res:
            results.append({'label': '%s : %s' % (word, ', '.join(sorted(labels))),
                            'value': word})
        self.w(json_dumps(results))

    def _build_results_for_search(self, rset, all_words):
        _ = self._cw._
        processed = defaultdict(list)
        words = []
        for word, etype, score in rset:
            if all_words:
                # test the hole phrase exists
                res = self.rqlexec('Any X ORDERBY FTIRANK(X) DESC WHERE X has_text %(text)s',
                                   {'text': '%s %s' % (all_words, word)})
            if not all_words or res:
                processed[word].append(_(etype))
                if word not in words:
                    words.append(word)
        return [(w, processed[w]) for w in words]

    def _build_results_for_facet(self, rset, rql, all_words):
        _ = self._cw._
        select = self._cw.vreg.parse(self._cw, rql).children[0]
        filtered_variable = facetbase.get_filtered_variable(select)
        processed = defaultdict(list)
        words = []
        for word, etype, score in rset:
            if word in processed:
                processed[word].append(_(etype))
                continue
            if all_words:
                # test the hole phrase exists
                select.save_state()
                try:
                    select.add_constant_restriction(filtered_variable,
                                                    'has_text',
                                                    '%s %s' % (all_words, word),
                                                    'String')
                    res = self.rqlexec(select.as_string())
                finally:
                    select.recover()
            if not all_words or res:
                processed[word].append(_(etype))
                words.append(word)
        return [(w, processed[w]) for w in words]


class AutocompleteSearchBox(SearchBox):
    formdef = '''
<form action="%(action)s" id="search_box" class="navbar-form" role="search">
  <input id="norql" autocomplete="off" type="text" accesskey="q" tabindex="%(tabindex1)s"
    title="search text" value="%(value)s" name="rql"
    class="search-query form-control" placeholder="%(searchlabel)s"/>
  <input type="hidden" name="__fromsearchbox" value="1" />
  <input type="hidden" name="subvid" value="tsearch" />
</form>
'''

    def render_body(self, w):
        self._cw.add_js(('jquery.ui.js', 'cubicweb.autocompletesearch.js'))
        self._cw.add_css(('jquery.ui.css',))
        url = self._cw.build_url('view', vid='search-autocomplete')
        minlength = 2
        limit = 20
        self._cw.add_onload('cw.search_autocomplete($("#norql"), "%s", %s, %s);'
                            % (url, minlength, limit))
        super(AutocompleteSearchBox, self).render_body(w)


class AutocompleteSearchWidget(facetbase.FacetStringWidget):

    def _render(self):
        """ has-hext facet widget with autocomplete"""
        req = self.facet._cw
        facetid = domid(make_uid(self.facet.__regid__))
        input_id = xml_escape('id-%s' % facetid)
        req.add_js(('jquery.ui.js', 'cubicweb.autocompletesearch.js'))
        req.add_css(('jquery.ui.css', ))
        url = req.build_url('view', vid='search-autocomplete')
        minlength = 2
        limit = 20
        req.add_onload('cw.search_autocomplete($("#%s"), "%s", %s, %s);' % (
            input_id, url, minlength, limit))
        w = self.w
        title = xml_escape(self.facet.title)
        w(u'<div id="%s" class="facet">\n' % facetid)
        cssclass = 'facetTitle'
        if self.facet.allow_hide:
            cssclass += ' hideFacetBody'
        w(u'<div class="%s" cubicweb:facetName="%s">%s</div>\n' %
            (cssclass, xml_escape(self.facet.__regid__), title))
        cssclass = 'facetBody'
        if not self.facet.start_unfolded:
            cssclass += ' hidden'
        w(u'<div class="%s">\n' % cssclass)
        w((u'<input autocomplete="off" id="%s" name="%s" '
            'type="text" value="%s" data-facet="1"/>\n') % (
            input_id, xml_escape(self.facet.__regid__),
            self.value or u''))
        w(u'</div>\n')
        w(u'</div>\n')


def registration_callback(vreg):
    try:
        import psycopg2
    except ImportError:
        psycopg2 = None
    if psycopg2:
        vreg.register_all(globals().values(), __name__, [])
        vreg.unregister(SearchBox)
        HasTextFacet.wdgclass = AutocompleteSearchWidget



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref test for OAI-PMH export"""

from collections import namedtuple
from datetime import date, datetime, timedelta
from functools import wraps
import time

from dateutil.parser import parse as parse_date
from lxml import etree
import pytz

from logilab.common import tempattr

from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref import isodate
from cubes.saem_ref.entities.oai import OAIComponent
from cubes.saem_ref.views.oai import OAIError, OAIRequest

import testutils


EntityInfo = namedtuple('EntityInfo', ['ark', 'url', 'eid'])


def publish_entity(entity):
    """Fire "publish" transition for `entity`."""
    entity.cw_adapt_to('IWorkflowable').fire_transition('publish')
    entity._cw.commit()


def deprecate_entity(entity):
    """Fire "deprecate" transition for `entity`."""
    entity.cw_adapt_to('IWorkflowable').fire_transition('deprecate')
    entity._cw.commit()


class OAITestMixin(object):

    def oai_component(self, cnx):
        """Return the "oai" component"""
        return self.vreg['components'].select('oai', cnx)


class OAIComponentTC(CubicWebTC, OAITestMixin):

    def test_registered(self):
        with self.admin_access.repo_cnx() as cnx:
            oai = self.oai_component(cnx)
            setspecs = oai.__setspecs__
            self.assertCountEqual(setspecs.keys(),
                                  ['agent', 'conceptscheme', 'profile', 'concept'])
            self.assertCountEqual(setspecs['agent'].keys(),
                                  ['kind', 'role'])
            self.assertCountEqual(setspecs['profile'].keys(),
                                  ['transferring_agent'])
            self.assertEqual(setspecs['conceptscheme'].keys(), [])

    def test_match_setspec(self):
        with self.admin_access.repo_cnx() as cnx:
            oai = self.oai_component(cnx)
            with self.assertRaises(ValueError) as cm:
                oai.match_setspec(u'cwuser')
            self.assertIn('no setspec matching cwuser', str(cm.exception))

    def test_setspecs(self):
        with self.admin_access.repo_cnx() as cnx:
            arch = testutils.agent(cnx, u'arch', archival_roles=[u'archival'])
            ark = testutils.agent(cnx, u'dep', archival_roles=[u'deposit'],
                                  archival_agent=arch).ark
            cnx.commit()
            scheme = setup_scheme(cnx, u'test', u'lab')
            publish_entity(scheme)
            expected = (
                'agent',
                'agent:kind:person', 'agent:kind:authority', 'agent:kind:family',
                'agent:role:producer', 'agent:role:deposit',
                'agent:role:archival', 'agent:role:control',
                'agent:role:enquirer', 'agent:role:seda-actor',
                'conceptscheme',
                'profile', 'profile:transferring_agent:%s' % ark,
                'concept', 'concept:in_scheme:%s' % scheme.ark
            )
            oai = self.oai_component(cnx)
            setspecs = [x[0] for x in oai.setspecs()]
            self.assertCountEqual(setspecs, expected)


def setup_agents(cnx):
    """Create three agents (alice, bob and draft), publish alice and bob but not draft, and return
    the (ark, url, eid) attributes of alice and bob.
    """
    alice = testutils.agent(cnx, u'alice', archival_roles=[u'producer'])
    cnx.commit()
    publish_entity(alice)
    bob = testutils.agent(cnx, u'bob', archival_roles=[u'enquirer'],
                          kind=u'family')
    cnx.commit()
    publish_entity(bob)
    testutils.agent(cnx, u'draft', kind=u'family',
                    archival_roles=[u'enquirer'])
    cnx.commit()
    return (
        EntityInfo(alice.ark, alice.absolute_url(), alice.eid),
        EntityInfo(bob.ark, bob.absolute_url(), bob.eid),
    )


def setup_scheme(cnx, title, *labels):
    """Return info new concept scheme"""
    scheme = cnx.create_entity('ConceptScheme', title=title)
    for label in labels:
        cnx.create_entity(
            'Label', label=label,
            label_of=cnx.create_entity('Concept', in_scheme=scheme))
    cnx.commit()
    return scheme


class OAIRequestTC(CubicWebTC):

    def test_rset_from_setspec(self):
        """Test `rset_from_setspec` function for exceptions."""
        with self.admin_access.repo_cnx() as cnx:
            return setup_agents(cnx)
        with self.admin_access.web_request() as req:
            for setspec, msg in [
                ('agent:zen', 'invalid set specifier agent:zen'),
                ('agent:role:control', (
                    'The combination of the values of the from, until, and '
                    'set arguments results in an empty list.')),
                ('conceptscheme:haha', 'invalid set specifier conceptscheme:haha'),
                ('scregneugneu', 'invalid set specifier scregneugneu'),
            ]:
                self.set_description(setspec)
                oairq = OAIRequest('http://testing.fr', setspec=setspec)
                with self.assertRaises(OAIError) as cm:
                    oairq.rset_from_setspec(req)
                self.assertEqual(cm.exception.errors, {'noRecordsMatch': msg})

    def test_rset_from_setspec_token(self):
        with self.admin_access.repo_cnx() as cnx:
            alice, bob = setup_agents(cnx)

        def oairq(**kwargs):
            return OAIRequest('http://testing.fr', setspec='agent', **kwargs)

        with self.admin_access.web_request() as req:
            with tempattr(OAIComponent, 'max_result_size', 1):
                rset, token = oairq().rset_from_setspec(req)
                self.assertEqual(len(rset), 1)
                self.assertEqual(token, str(bob.eid))
                rset, token = oairq(resumption_token=str(bob.eid)).rset_from_setspec(req)
                self.assertEqual(len(rset), 1)
                self.assertIsNone(token)
            # All results are returned, the resumptionToken is None.
            rset, token = oairq().rset_from_setspec(req)
            self.assertEqual(len(rset), 2)
            self.assertIsNone(token)
            # and this, even if the user specified an input token.
            rset, token = oairq(resumption_token='1').rset_from_setspec(req)
            self.assertEqual(len(rset), 2)
            self.assertIsNone(token)


def no_validate_xml(method):
    """Disable XML schema validation, often because the underlying metadata
    part of the response (RDF, XSD) is not validable (or we don't know how to
    do it).
    """
    @wraps(method)
    def wrapper(self):
        self._validate_xml = False
        self._debug_xml = True
        return method(self)
    return wrapper


class OAIPMHViewsTC(CubicWebTC, OAITestMixin, testutils.XmlTestMixin):

    _validate_xml = True
    _debug_xml = False

    def oai_request(self, req, **formparams):
        req.form.update(formparams)
        out = self.app_handle_request(req, 'oai')
        if self._validate_xml:
            self.assertXmlValid(out, self.datapath('OAI-PMH.xsd'), debug=self._debug_xml)
        return out

    def _setup_agents(self, **kwargs):
        with self.admin_access.repo_cnx() as cnx:
            return setup_agents(cnx, **kwargs)

    def test_xml_attributes_and_namespaces(self):
        """Check XML attributes and namespace declaration of the response."""
        with self.admin_access.web_request() as req:
            # simple request, generating an error but enough to get a properly
            # formatter response.
            result = self.oai_request(req)
            xml = etree.fromstring(result)
            nsmap = {None: 'http://www.openarchives.org/OAI/2.0/',
                     'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}
            self.assertEqual(xml.nsmap, nsmap)
            attrib = {
                '{http://www.w3.org/2001/XMLSchema-instance}schemaLocation': ' '.join(
                    ['http://www.openarchives.org/OAI/2.0/',
                     'http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd']),
            }
            self.assertEqual(xml.attrib, attrib)

    def test_responsedate(self):
        """Check <responseDate>"""
        with self.admin_access.web_request() as req:
            # Simple request, generating an error but enough to get a properly
            # formatter response.
            now = datetime.utcnow()
            result = self.oai_request(req)
        # Check tzinfo abbreviation is present (Z for UTC).
        self.assertIn('<responseDate>{0}</responseDate>'.format(isodate(now)),
                      result)

    def _setup_agents_deleted(self):
        alice, bob = self._setup_agents()
        with self.admin_access.repo_cnx() as cnx:
            entity = cnx.entity_from_eid(bob.eid)
            # ensure transition date is greater by at least 1s
            time.sleep(1)
            date_before = entity.modification_date.replace(tzinfo=pytz.utc)
            deprecate_entity(entity)
        return alice, bob, date_before

    def _setup_schemes(self):
        with self.admin_access.repo_cnx() as cnx:
            pscheme = setup_scheme(cnx, u'public',
                                   u'public label', u'other public label')
            publish_entity(pscheme)
            # draft ConceptScheme
            dscheme = setup_scheme(cnx, u'draft',
                                   u'draft label', u'other draft label')
            return (
                EntityInfo(pscheme.ark, pscheme.absolute_url(), pscheme.eid),
                EntityInfo(dscheme.ark, dscheme.absolute_url(), dscheme.eid),
            )

    def test_noverb(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req)
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn('<error code="badVerb">no verb specified</error',
                      result)

    def test_badverb(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='coucou')
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn('<error code="badVerb">illegal verb "coucou"</error>',
                      result)

    # some of our setspecs (those with an ARK inside are not valid)
    @no_validate_xml
    def test_listsets(self):
        pscheme, dscheme = self._setup_schemes()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListSets')
        self.assertIn('<request verb="ListSets">http://testing.fr/cubicweb/oai</request>',
                      result)
        for spec in ('agent', 'conceptscheme', 'profile', 'concept'):
            self.assertIn('<setSpec>{0}</setSpec>'.format(spec),
                          result)
        self.assertIn(('<setSpec>agent:role:deposit</setSpec>'
                       '<setName>An agent with deposit archival role</setName>'),
                      result)
        self.assertIn('<setSpec>concept:in_scheme:{0}</setSpec>'.format(pscheme.ark),
                      result)
        self.assertNotIn(dscheme.ark, result)

    def test_listidentifiers_noset(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers')
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn(('<error code="badArgument">'
                       'ListIdentifiers verb requires "set" restriction'
                       '</error>'),
                      result)

    def _setup_profile(self):
        """Create a "published" SEDAProfile and return its identifier"""
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            cnx.commit()
            publish_entity(profile)
            return profile.ark

    def test_resumptiontoken_listidentifiers(self):
        alice, bob = self._setup_agents()

        def check_request(expected_arks, token=None):
            with self.admin_access.web_request() as req:
                result = self.oai_request(
                    req, verb='ListIdentifiers', set='agent',
                    resumptionToken=token)
                # Ensure there are as many <identifier> tag than expected items.
                self.assertEqual(result.count('<identifier>'), len(expected_arks))
                for ark in expected_arks:
                    self.assertIn(
                        '<identifier>{0}</identifier>'.format(ark), result)
                return result

        self.set_description('less results than pagination limit')
        result = check_request([alice.ark, bob.ark])
        # All results returned, no resumptionToken is request -> no
        # resumptionToken in response.
        self.assertNotIn('resumptionToken', result)
        with tempattr(OAIComponent, 'max_result_size', 1):
            self.set_description('max_result_size = 1, 1st batch')
            result = check_request([alice.ark])
            self.set_description('token + expirationDate')
            root = etree.fromstring(result)
            token = root.find('.//{%s}resumptionToken' % root.nsmap[None])
            date = parse_date(token.attrib['expirationDate'])
            self.assertLess(datetime.now(pytz.utc) + timedelta(hours=1) - date,
                            timedelta(seconds=1))
            self.assertEqual(token.text, str(bob.eid))
            self.set_description('max_result_size = 1, 2nd batch')
            result = check_request([bob.ark], token=str(bob.eid))
            self.set_description('max_result_size = 1, no more result')
            # No more results -> empty resumptionToken.
            self.assertIn('<resumptionToken/>', result)
            # Empty result, probably a badResumptionToken.
            result = check_request([], token=str(bob.eid + 1))
            self.assertIn('badResumptionToken', result)

    def test_listidentifiers(self):
        alice, bob = [x.ark for x in self._setup_agents()]
        profile_id = self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent')
            self.assertIn('<identifier>{0}</identifier>'.format(bob), result)
            self.assertIn('<identifier>{0}</identifier>'.format(alice), result)
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            self.assertNotIn(draft, result)
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:role:producer')
            self.assertIn('<identifier>{0}</identifier>'.format(alice),
                          result)
            self.assertNotIn('<identifier>{0}</identifier>'.format(bob),
                             result)
            result = self.oai_request(req, verb='ListIdentifiers', set='agent:kind:person')
            self.assertIn('<identifier>{0}</identifier>'.format(alice), result)
            self.assertNotIn('<identifier>{0}</identifier>'.format(bob), result)
            result = self.oai_request(req, verb='ListIdentifiers', set='conceptscheme')
            result = self.oai_request(req, verb='ListIdentifiers', set='profile')
            self.assertIn('<identifier>{0}</identifier>'.format(profile_id), result)

    def test_listidentifiers_setspec_wfstate(self):
        """Check workflow state condition is statisfied with a multi-level
        setspec `agent:role:enquirer`.
        """
        alice, bob = [x.ark for x in self._setup_agents()]
        profile_id = self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:role:enquirer')
            self.assertIn('<identifier>{0}</identifier>'.format(bob), result)
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            self.assertNotIn(draft, result)

    def test_listidentifiers_deleted(self):
        """Check <header> element (status attribute, datestamp) for deleted
        records in ListIdentifiers response.
        """
        alice, bob, date_before = self._setup_agents_deleted()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent')
            self.assertIn(
                '<identifier>{0}</identifier>'.format(bob.ark), result)
            self.assertIn('<header status="deleted">', result)
            root = etree.fromstring(result)
            ns = root.nsmap[None]
            # only one <header status="..."> element
            header, = root.findall('.//{%s}header[@status]' % ns)
            datestamp = header.find('{%s}datestamp' % ns).text
            date = parse_date(datestamp)
            self.assertGreater(date, date_before.replace(microsecond=0))

    def test_listrecords_noset(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords')
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn(('<error code="badArgument">'
                       'ListRecords verb requires "set" restriction'
                       '</error>'),
                      result)

    @no_validate_xml
    def test_listrecords(self):
        alice, bob = self._setup_agents()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords', set='agent')
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(bob.url), result)
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(alice.url), result)
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            self.assertNotIn(draft, result)
            result = self.oai_request(
                req, verb='ListRecords', set='agent:role:producer')
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(alice.url), result)
            self.assertNotIn(bob.url, result)
            result = self.oai_request(req, verb='ListRecords', set='agent:kind:person')
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(alice.url), result)
            self.assertNotIn(bob.url, result)

    @no_validate_xml
    def test_listrecord_deleted(self):
        """Check absence of <metadata> or <about> elements for deleted
        records in ListRecords response.
        """
        alice, bob, _ = self._setup_agents_deleted()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords', set='agent')
            # no metadata for bob
            self.assertNotIn(
                '<rdf:Description rdf:about="{0}">'.format(bob.url), result)
            # but record present
            self.assertIn('<identifier>{0}</identifier>'.format(bob.ark), result)
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(alice.url), result)

    @no_validate_xml
    def test_listrecords_profile(self):
        self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords', set='profile')
        # Just ensure a SEDA profile (xsd) is present in response.
        self.assertIn('<metadata><xsd:schema', result)
        self.assertIn('fr:gouv:ae:archive:draft:standard_echange_v0.2', result)

    @no_validate_xml
    def test_listidentifiers_profile_setspec(self):
        """Check setspec restriction for SEDAProfile."""
        profile_id = self._setup_profile()
        profile2_id = self._setup_profile()
        with self.admin_access.web_request() as req:
            transferring_agent_ark = req.execute(
                'Any K WHERE P ark "%s", P seda_transferring_agent A, A ark K'
                % profile_id)[0][0]
            result = self.oai_request(
                req, verb='ListIdentifiers',
                set='profile:transferring_agent:' + transferring_agent_ark)
            self.assertIn('<identifier>{0}</identifier>'.format(profile_id), result)
            self.assertNotIn(profile2_id, result)
        # Check that without a setspec, we get all profiles.
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='profile')
            self.assertIn('<identifier>{0}</identifier>'.format(profile_id), result)
            self.assertIn('<identifier>{0}</identifier>'.format(profile2_id), result)

    @no_validate_xml
    def test_getrecord_profile(self):
        profile_id = self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='GetRecord')
            self.assertIn('<error code="badArgument">', result)
            result = self.oai_request(req, verb='GetRecord', identifier='999')
            self.assertIn('<error code="idDoesNotExist"', result)
            result = self.oai_request(req, verb='GetRecord', identifier=profile_id)
            self.assertIn('<metadata><xsd:schema', result)
            self.assertIn('fr:gouv:ae:archive:draft:standard_echange_v0.2', result)

    @no_validate_xml
    def test_getrecord_agent(self):
        self._debug_xml = True
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.agent(cnx, u'toto')
            cnx.commit()
            publish_entity(agent)
            ark = agent.ark
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='GetRecord', identifier=ark)
            self.assertIn('name>toto</', result)
            self.assertIn('<rdf:type rdf:resource="http://xmlns.com/foaf/0.1/Person"/>',
                          result)

    def test_getrecord_deleted(self):
        """Check absence of <metadata> or <about> elements for deleted
        records in GetRecord response.
        """
        alice, bob, _ = self._setup_agents_deleted()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='GetRecord', identifier=bob.ark)
            # record is present
            self.assertIn('<identifier>{0}</identifier>'.format(bob.ark), result)
            # but not data
            self.assertNotIn('metadata', result)
            self.assertNotIn('rdf', result)

    def test_getrecord_draft(self):
        self._setup_agents()
        with self.admin_access.web_request() as req:
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            result = self.oai_request(req, verb='GetRecord', identifier=draft)
            self.assertIn('idDoesNotExist', result)

    def test_from_until_error(self):
        with self.admin_access.web_request() as req:
            dates = {'from': date(2013, 1, 2).isoformat(),
                     'until': date(2011, 1, 2).isoformat()}
            result = self.oai_request(req, verb='ListIdentifiers', set='agent',
                                      **dates)
            self.assertIn('<error code="badArgument">the from argument', result)

    def test_from_until_listidentifiers(self):
        alice, bob = self._setup_agents()
        with self.admin_access.repo_cnx() as cnx:
            alice_date = cnx.entity_from_eid(alice.eid).modification_date
            bob_date = alice_date + timedelta(seconds=2)
            cnx.entity_from_eid(bob.eid).cw_set(modification_date=bob_date)
            cnx.commit()
        in_between_date = isodate(alice_date + timedelta(seconds=1))
        self.set_description('from only')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent',
                                      **{'from': in_between_date})
            self.assertIn(bob.ark, result)
            self.assertNotIn(alice.ark, result)
        self.set_description('until only')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent',
                                      **{'until': in_between_date})
            self.assertNotIn(bob.ark, result)
            self.assertIn(alice.ark, result)
        self.set_description('from and until')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent',
                                      **{'from': '1970-12-1',
                                         'until': '2099-01-02'})
            self.assertIn(bob.ark, result)
            self.assertIn(alice.ark, result)
        self.set_description('no record match')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent',
                                      **{'until': '1970-12-31'})
            self.assertIn('<error code="noRecordsMatch">The combi', result)

    def test_from_until_listidentifiers_setspec(self):
        """Check from/until restriction with a multi-level setspec
        `agent:kind:family`
        """
        alice, bob = self._setup_agents()
        with self.admin_access.repo_cnx() as cnx:
            bob_date = cnx.entity_from_eid(bob.eid).modification_date
            marcel = cnx.find('Agent', name=u'draft').one()
            publish_entity(marcel)
            marcel_date = bob_date + timedelta(seconds=2)
            marcel.cw_set(name=u'marcel', modification_date=marcel_date)
            marcel_ark = marcel.ark
            cnx.commit()
        in_between_date = isodate(bob_date + timedelta(seconds=1))
        self.set_description('from only')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:kind:family',
                                      **{'from': in_between_date})
            self.assertNotIn(bob.ark, result)
            self.assertIn(marcel_ark, result)
            self.assertNotIn(alice.ark, result)
        self.set_description('until only')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:kind:family',
                                      **{'until': in_between_date})
            self.assertIn(bob.ark, result)
            self.assertNotIn(marcel_ark, result)
            self.assertNotIn(alice.ark, result)
        self.set_description('from and until')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:kind:family',
                                      **{'from': '1970-12-1',
                                         'until': '2099-01-02'})
            self.assertIn(bob.ark, result)
            self.assertIn(marcel_ark, result)
            self.assertNotIn(alice.ark, result)
        self.set_description('no record match')
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:kind:family',
                                      **{'until': '1970-12-31'})
            self.assertIn('<error code="noRecordsMatch">The combi', result)


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref custom views for skos entities"""

from cubicweb import NoSelectableObject, tags
from cubicweb.view import EntityView, AnyRsetView
from cubicweb.predicates import is_instance
from cubicweb.web.views import baseviews, tabs, treeview, uicfg

from cubes.skos import views as skos
from cubes.saem_ref.views import ImportEntityComponent

_ = unicode

afs = uicfg.autoform_section


_('Import ConceptScheme')  # generate message used by the import component


class SKOSImportComponent(ImportEntityComponent):
    """Component with a link to import a concept scheme from a SKOS file."""
    __select__ = ImportEntityComponent.__select__ & is_instance('ConceptScheme')

    @property
    def import_url(self):
        return self._cw.build_url('add/skossource')


class SKOSActivityTabView(tabs.TabsMixin, EntityView):
    """Life-cycle tab for concept and concept scheme."""
    __regid__ = 'saem_ref_vocabulary_lifecycle'
    __select__ = is_instance('ConceptScheme', 'Concept')

    def entity_call(self, entity):
        try:
            self._cw.view('prov.activity-generated', entity=entity, w=self.w, tabid=self.__regid__)
        except NoSelectableObject:
            self.w(u'<div>%s</div>\n' % self._cw.__('No result matching query'))


afs.tag_object_of(('*', 'related_concept_scheme', 'ConceptScheme'), 'main', 'hidden')
afs.tag_object_of(('*', 'seda_keyword_scheme', 'ConceptScheme'), 'main', 'hidden')

skos.ConceptSchemePrimaryView.tabs.append(_('saem_ref_vocabulary_lifecycle'))
skos.ConceptPrimaryView.tabs.append(_('saem_ref_vocabulary_lifecycle'))


class ConceptSchemeSameETypeListView(baseviews.SameETypeListView):
    """Override SameETypeListView to display a search action on top"""
    __select__ = baseviews.SameETypeListView.__select__ & is_instance('ConceptScheme')

    def call(self, **kwargs):
        self.search_link()
        super(ConceptSchemeSameETypeListView, self).call(**kwargs)

    def search_link(self):
        """Render a link to search for Concepts within selected ConceptSchemes"""
        rql = 'Any X WHERE X in_scheme S, S eid '
        if len(self.cw_rset) > 1:
            rql += 'IN ({0})'.format(', '.join(str(x[0]) for x in self.cw_rset.rows))
        else:
            rql += str(self.cw_rset[0][0])
        href = self._cw.build_url(rql=rql)
        title = self._cw._('search for concepts of concept schemes')
        self.w(u'<a class="btn btn-default pull-right" href="{href}">{title}</a>'.format(
            href=href, title=title))


class ConceptTreeView(treeview.TreeView):
    """`treeview` for Concept entities."""
    __select__ = treeview.TreeView.__select__ & is_instance('Concept')
    subvid = 'skos.concept-popover'


class ConceptPopoverView(AnyRsetView):
    """Popover view for Concept, to be displayed within treeview"""
    __regid__ = 'skos.concept-popover'
    __select__ = is_instance('Concept')

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        self._cw.add_onload("$('#{eid}').popover();".format(eid=entity.eid))
        title = entity.dc_title()
        content = u'<dl>'
        for attr in ('definition', 'example'):
            if getattr(entity, attr):
                content += u'<dt>{label}</dt><dd>{value}</dd>'.format(
                    label=self._cw._(attr),
                    value=entity.printable_value(attr, format=u'text/html'))
        content += u'</dl>'
        content += tags.a(self._cw._('view'), href=entity.absolute_url())
        data = {'data-title': title,
                'data-content': content,
                'data-html': 'true',
                'data-trigger': 'focus'}
        self.w(tags.a(title, href='javascript:$.noop();', id=str(entity.eid), **data))


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)
    from cubes.skos.views import ImportConceptSchemeAction
    vreg.unregister(ImportConceptSchemeAction)
    from cubicweb.web.views import cwsources
    vreg.unregister(cwsources.ManageSourcesAction)



# coding: utf-8
"""cubicweb-saem_ref unit tests for views"""

from logilab.common import flatten

from cubicweb.devtools import PostgresApptestConfiguration, startpgcluster, stoppgcluster
from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref.views.autocompletesearch import get_results_from_query


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


def sql(cnx, string, param=None):
    return cnx.system_sql(string, param).fetchall()


class AutocompleteSearchTC(CubicWebTC):
    configcls = PostgresApptestConfiguration

    def test_syncro_concepts(self):
        with self.admin_access.client_cnx() as cnx:
            self.assertFalse(sql(cnx, 'SELECT * from words WHERE word=%(w)s;',
                                 {'w': u'déclaration'}))
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus')
            concept = scheme.add_concept(label=u'Déclaration universelle', language_code=u'fr',
                                         definition=u"Déclaration des droits de l'homme")
            scheme.add_concept(label=u'The Universal Declaration', language_code=u'fr',
                               definition=u'Declaration of human rights and citizen')
            cnx.commit()
            all_words = flatten(sql(cnx, 'SELECT word FROM words'))
            for word in (u'responsabilit\xe9', u'responsable', u'responsables'):
                self.assertIn(word, all_words)
            self.assertEqual(get_results_from_query(cnx, {'q': u'déclaration'}),
                             [(u'd\xe9claration', u'Concept', 1.0),
                              (u'declaration', u'Concept', 0.6)])
            self.assertEqual(get_results_from_query(cnx, {'q': u'statisti'}),
                             [(u'statistica', u'Concept', 0.666667),
                              (u'statistique', u'Concept', 0.615385),
                              (u'statistiques', u'Concept', 0.571429)])
            self.assertEqual(get_results_from_query(cnx, {'q': u'rspons'}),
                             [])
            concept.preferred_label[0].cw_set(label=u'Loi renseignement')
            concept.cw_set(definition=u'''L'Assemblée et le Sénat se sont mis d'accord, mardi 16
            juin, sur une version définitive de la loi renseignement. Un texte voté dans une sorte
            d'indifférence générale et qui dote la France d'une des lois les plus intrusives
            d’Europe.''')
            cnx.commit()
            self.assertEqual(get_results_from_query(cnx, {'q': u'déclaration'}),
                             [(u'declaration', u'Concept', 0.6)])

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



drop_entity_type('MirrorEntity')
drop_entity_type('AlfrescoConnector')
drop_entity_type('AsalaeConnector')

for ertype in ('Label', 'ConceptScheme', 'in_scheme', 'broader_concept', 'label_of',
               'SEDAProfile'):
    sync_schema_props_perms(ertype, syncrdefs=False, syncprops=False)

print( schema['vocabulary_source'].rdefs)
print ('equivalent_concept' in schema)

with session.deny_all_hooks_but('metadata', 'syncschema'):
    rename_relation_type('vocabulary_source', 'equivalent_concept', force=True)
add_relation_type('vocabulary_source')

rql('SET X vocabulary_source SC WHERE X equivalent_concept C, C in_scheme SC')
commit()

add_entity_type('Citation')
add_entity_type('Occupation')




# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref site customizations"""

import pytz

from logilab.common.date import ustrftime
from logilab.common.decorators import monkeypatch

from yams.constraints import OPERATORS, BoundaryConstraint, Attribute

from cubicweb import cwvreg
from cubicweb.cwconfig import register_persistent_options
from cubicweb.uilib import PRINTERS

from cubes.skos import rdfio
from cubes.skos.ccplugin import ImportSkosData

from cubes.saem_ref import cwuri_url
from cubes.saem_ref.sobjects.skos import SAEMMetadataGenerator
# this import is needed to take account of pg_trgm monkeypatches
# while executing cubicweb-ctl commands (db-rebuild-fti)
from cubes.saem_ref import pg_trgm  # noqa pylint: disable=unused-import

_ = unicode


@monkeypatch(rdfio.RDFGraphGenerator, methodname='equivalent_concept_uris')
@staticmethod
def equivalent_concept_uris(entity):
    yield cwuri_url(entity)


# deactivate date-format and datetime-format cw properties. This is because we do some advanced date
# manipulation such as allowing partial date and this is not generic enough to allow arbitrary
# setting of date and time formats

base_user_property_keys = cwvreg.CWRegistryStore.user_property_keys


@monkeypatch(cwvreg.CWRegistryStore)
def user_property_keys(self, withsitewide=False):
    props = base_user_property_keys(self, withsitewide)
    return [prop for prop in props if prop not in ('ui.date-format', 'ui.datetime-format')]

# customize display of TZDatetime

register_persistent_options((
    ('timezone',
     {'type': 'choice',
      'choices': pytz.common_timezones,
      'default': 'Europe/Paris',
      'help': _('timezone in which time should be displayed'),
      'group': 'ui', 'sitewide': True,
      }),
))


def print_tzdatetime_local(value, req, *args, **kwargs):
    tz = pytz.timezone(req.property_value('ui.timezone'))
    value = value.replace(tzinfo=pytz.utc).astimezone(tz)
    return ustrftime(value, req.property_value('ui.datetime-format'))

PRINTERS['TZDatetime'] = print_tzdatetime_local


# configure c-c skos-import command's factories to use with proper metadata generator ##############

def _massive_store_factory(cnx):
    from cubes.saem_ref.backports.massive_store import MassiveObjectStore
    return MassiveObjectStore(cnx, metagen=SAEMMetadataGenerator(cnx))


def _nohook_store_factory(cnx):
    from cubes.saem_ref.backports.stores import NoHooRQLObjectStore
    return NoHooRQLObjectStore(cnx, metagen=SAEMMetadataGenerator(cnx))


ImportSkosData.cw_store_factories['massive'] = _massive_store_factory
ImportSkosData.cw_store_factories['nohook'] = _nohook_store_factory


####################################################################################################
# temporary monkey-patches #########################################################################
####################################################################################################

# support for custom message on boundary constraint (https://www.logilab.org/ticket/288874)

@monkeypatch(BoundaryConstraint)
def __init__(self, op, boundary=None, msg=None):
    assert op in OPERATORS, op
    self.msg = msg
    self.operator = op
    self.boundary = boundary


@monkeypatch(BoundaryConstraint)
def failed_message(self, key, value):
    if self.msg:
        return self.msg, {}
    boundary = self.boundary
    if isinstance(boundary, Attribute):
        boundary = boundary.attr
    return "value %(KEY-value)s must be %(KEY-op)s %(KEY-boundary)s", {
        key + '-value': value,
        key + '-op': self.operator,
        key + '-boundary': boundary}


@monkeypatch(BoundaryConstraint)
def serialize(self):
    """simple text serialization"""
    value = u'%s %s' % (self.operator, self.boundary)
    if self.msg:
        value += '\n' + self.msg
    return value


@monkeypatch(BoundaryConstraint, methodname='deserialize')
@classmethod
def deserialize(cls, value):
    """simple text deserialization"""
    try:
        value, msg = value.split('\n', 1)
    except ValueError:
        msg = None
    op, boundary = value.split(' ', 1)
    return cls(op, eval(boundary), msg or None)


# monkey-patch for https://www.cubicweb.org/ticket/5352619 #########################################
# other part in views/patches.py

from cubicweb import neg_role  # noqa
from cubicweb.web import formfields  # noqa

orig_guess_field = formfields.guess_field


@monkeypatch(formfields)
def guess_field(eschema, rschema, role='subject', req=None, **kwargs):
    rdef = eschema.rdef(rschema, role)
    card = rdef.role_cardinality(role)
    composite = getattr(rdef, 'composite', None)
    # don't mark composite relation as required, we want the composite element to be removed when
    # not linked to its parent
    kwargs.setdefault('required', card in '1+' and composite != neg_role(role))
    return orig_guess_field(eschema, rschema, role, req, **kwargs)


# set eid on eschema during instance creation (https://www.cubicweb.org/ticket/10450092) ###########

from cubicweb import server  # noqa

orig_initialize_schema = server.initialize_schema


@monkeypatch(server)
def initialize_schema(config, schema, mhandler, event='create'):
    schema = mhandler.repo.schema
    orig_initialize_schema(config, schema, mhandler, event)


# backport warning instead of failure when some extentity has several value for an entity ##########
# to be removed once we depend on cw 3.23

from cubicweb.dataimport import importer  # noqa


@monkeypatch(importer.ExtEntity)
def prepare(self, schema, import_log):
    """Prepare an external entity for later insertion:

    * ensure attributes and inlined relations have a single value
    * turn set([value]) into value and remove key associated to empty set
    * remove non inlined relations and return them as a [(e1key, relation, e2key)] list

    The schema and import_log into which errors or warning may be emitted are given as
    arguments.

    Return a list of non inlined relations that may be inserted later, each relations defined by
    a 3-tuple (subject extid, relation type, object extid).

    Take care the importer may call this method several times.
    """
    assert self._schema is None, 'prepare() has already been called for %s' % self
    self._schema = schema
    eschema = schema.eschema(self.etype)
    deferred = []
    entity_dict = self.values
    for key, rtype, role in self.iter_rdefs():
        rschema = schema.rschema(rtype)
        if rschema.final or (rschema.inlined and role == 'subject'):
            if len(entity_dict[rtype]) > 1:
                values = ', '.join(repr(v) for v in entity_dict[rtype])
                import_log.record_warning(
                    "more than one value for attribute %r, only one will be kept: %s"
                    % (rtype, values), path=self.extid)
            if entity_dict[key]:
                entity_dict[rtype] = entity_dict[key].pop()
                if key != rtype:
                    del entity_dict[key]
                if (rschema.final and eschema.has_metadata(rtype, 'format')
                        and not rtype + '_format' in entity_dict):
                    entity_dict[rtype + '_format'] = u'text/plain'
            else:
                del entity_dict[key]
        else:
            for target_extid in entity_dict.pop(key):
                if role == 'subject':
                    deferred.append((self.extid, rtype, target_extid))
                else:
                    deferred.append((target_extid, rtype, self.extid))
    return deferred


@monkeypatch(importer.ExtEntitiesImporter)
def iter_ext_entities(self, ext_entities, deferred, queue):
    """Yield external entities in an order which attempts to satisfy
    schema constraints (inlined / cardinality) and to optimize the import.
    """
    schema = self.schema
    extid2eid = self.extid2eid
    for ext_entity in ext_entities:
        # check data in the transitional representation and prepare it for
        # later insertion in the database
        for subject_uri, rtype, object_uri in ext_entity.prepare(schema, self.import_log):
            deferred.setdefault(rtype, set()).add((subject_uri, object_uri))
        if not ext_entity.is_ready(extid2eid):
            queue.setdefault(ext_entity.etype, []).append(ext_entity)
            continue
        yield ext_entity
        # check for some entities in the queue that may now be ready. We'll have to restart
        # search for ready entities until no one is generated
        new = True
        while new:
            new = False
            for etype in self.etypes_order_hint:
                if etype in queue:
                    new_queue = []
                    for ext_entity in queue[etype]:
                        if ext_entity.is_ready(extid2eid):
                            yield ext_entity
                            # may unlock entity previously handled within this loop
                            new = True
                        else:
                            new_queue.append(ext_entity)
                    if new_queue:
                        queue[etype][:] = new_queue
                    else:
                        del queue[etype]



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import copy
from collections import defaultdict
from io import StringIO
from itertools import chain
from base64 import b64encode

from six.moves import range

from yams.constraints import SizeConstraint

from cubicweb.schema import PURE_VIRTUAL_RTYPES
from cubicweb.server.schema2sql import rschema_has_table
from cubicweb.dataimport import stores, pgstore


class MassiveObjectStore(stores.RQLObjectStore):
    """
    Store for massive import of data, with delayed insertion of meta data.

    WARNINGS:

    - This store may be only used with PostgreSQL for now, as it relies
      on the COPY FROM method, and on specific PostgreSQL tables to get all
      the indexes.

    - This store can only insert relations that are not inlined (i.e.,
      which do *not* have inlined=True in their definition in the schema), unless those are
      specified as entity attributes.

    It should be used as follows:

       store = MassiveObjectStore(cnx)
       store.init_rtype_table('Person', 'lives_in', 'Location')
       ...

       store.prepare_insert_entity('Person', subj_iid_attribute=person_iid, ...)
       store.prepare_insert_entity('Location', obj_iid_attribute=location_iid, ...)
       ...

       # subj_iid_attribute and obj_iid_attribute are argument names
       # chosen by the user (e.g. "cwuri"). These names can be identical.
       # person_iid and location_iid are unique IDs and depend on the data
       # (e.g URI).
       store.flush()
       store.relate_by_iid(person_iid, 'lives_in', location_iid)
       # For example:
       store.prepare_insert_entity('Person',
                                   cwuri='http://dbpedia.org/toto',
                                   name='Toto')
       store.prepare_insert_entity('Location',
                                   uri='http://geonames.org/11111',
                                   name='Somewhere')
       store.flush()
       store.relate_by_iid('http://dbpedia.org/toto',
                           'lives_in',
                           'http://geonames.org/11111')
       # Finally
       store.convert_relations('Person', 'lives_in', 'Location',
                               'subj_iid_attribute', 'obj_iid_attribute')
       # For the previous example:
       store.convert_relations('Person', 'lives_in', 'Location', 'cwuri', 'uri')
       ...
       store.commit()
       store.finish()
    """
    # max size of the iid, used to create the iid_eid conversion table
    iid_maxsize = 1024

    def __init__(self, cnx,
                 on_commit_callback=None, on_rollback_callback=None,
                 slave_mode=False,
                 eids_seq_range=10000,
                 metagen=None):
        """ Create a MassiveObject store, with the following attributes:

        - cnx: CubicWeb cnx
        - eids_seq_range: size of eid range reserved by the store for each batch
        """
        super(MassiveObjectStore, self).__init__(cnx)
        self.on_commit_callback = on_commit_callback
        self.on_rollback_callback = on_rollback_callback
        self.slave_mode = slave_mode
        self.eids_seq_range = eids_seq_range
        if metagen is None:
            metagen = stores.MetadataGenerator(cnx)
        self.metagen = metagen

        self.logger = logging.getLogger('dataimport.massive_store')
        self.sql = cnx.system_sql
        self.schema = self._cnx.vreg.schema
        self.default_values = get_default_values(self.schema)
        self.get_next_eid = lambda g=self._get_eid_gen(): next(g)
        self._dbh = PGHelper(cnx)

        cnx.read_security = False
        cnx.write_security = False

        self._data_entities = defaultdict(list)
        self._data_relations = defaultdict(list)
        self._initialized = set()
        # uri handling
        self._data_uri_relations = defaultdict(list)
        # etypes for which we have a uri_eid_%(etype)s table
        self._init_uri_eid = set()
        # etypes for which we have a uri_eid_%(e)s_idx index
        self._uri_eid_inserted = set()
        # set of rtypes for which we have a %(rtype)s_relation_iid_tmp table
        self._uri_rtypes = set()

        if not self.slave_mode:
            # drop constraint and metadata table, they will be recreated when self.finish() will be
            # called
            self._drop_all_constraints()
            self._drop_metatables_constraints()

    def _get_eid_gen(self):
        """ Function getting the next eid. This is done by preselecting
        a given number of eids from the 'entities_id_seq', and then
        storing them"""
        while True:
            last_eid = self._cnx.repo.system_source.create_eid(self._cnx, self.eids_seq_range)
            for eid in range(last_eid - self.eids_seq_range + 1, last_eid + 1):
                yield eid

    # URI related things #######################################################

    def init_rtype_table(self, etype_from, rtype, etype_to):
        """ Build temporary table for standard rtype """
        # Create an uri_eid table for each etype for a better control of which etype is concerned by
        # a particular possibly multivalued relation.
        for etype in (etype_from, etype_to):
            if etype and etype not in self._init_uri_eid:
                self._init_uri_eid.add(etype)
                self.sql('CREATE TABLE IF NOT EXISTS uri_eid_%(e)s'
                         '(uri character varying(%(size)s), eid integer)'
                         % {'e': etype.lower(), 'size': self.iid_maxsize})
        if rtype not in self._uri_rtypes:
            # Create the temporary table
            if not self.schema.rschema(rtype).inlined:
                self.sql('CREATE TABLE IF NOT EXISTS %(r)s_relation_iid_tmp'
                         '(uri_from character varying(%(s)s), uri_to character varying(%(s)s))'
                         % {'r': rtype, 's': self.iid_maxsize})
                self._uri_rtypes.add(rtype)
            else:
                self.logger.warning("inlined relation %s: cannot insert it", rtype)

    def relate_by_iid(self, iid_from, rtype, iid_to):
        """Add new relation based on the internal id (iid)
        of the entities (not the eid)"""
        # Push data
        if isinstance(iid_from, unicode):
            iid_from = iid_from.encode('utf-8')
        if isinstance(iid_to, unicode):
            iid_to = iid_to.encode('utf-8')
        self._data_uri_relations[rtype].append({'uri_from': iid_from, 'uri_to': iid_to})

    def flush_relations(self):
        """ Flush the relations data
        """
        for rtype, data in self._data_uri_relations.items():
            if not data:
                self.logger.info('No data for rtype %s', rtype)
            buf = StringIO('\n'.join(['%(uri_from)s\t%(uri_to)s' % d for d in data]))
            if not buf:
                self.logger.info('Empty Buffer for rtype %s', rtype)
                continue
            cursor = self._cnx.cnxset.cu
            if not self.schema.rschema(rtype).inlined:
                cursor.copy_from(buf, '%s_relation_iid_tmp' % rtype.lower(),
                                 null='NULL', columns=('uri_from', 'uri_to'))
            else:
                self.logger.warning("inlined relation %s: cannot insert it", rtype)
            buf.close()
            # Clear data cache
            self._data_uri_relations[rtype] = []

    def fill_uri_eid_table(self, etype, uri_label):
        """ Fill the uri_eid table
        """
        if etype not in self._uri_eid_inserted:
            self._uri_eid_inserted.add(etype)
            self.logger.info('Fill uri_eid for etype %s', etype)
            self.sql('INSERT INTO uri_eid_%(e)s SELECT cw_%(l)s, cw_eid FROM cw_%(e)s'
                     % {'l': uri_label, 'e': etype.lower()})
            self.sql('CREATE INDEX uri_eid_%(e)s_idx ON uri_eid_%(e)s(uri)'
                     % {'e': etype.lower()})

    def convert_relations(self, etype_from, rtype, etype_to,
                          uri_label_from='cwuri', uri_label_to='cwuri'):
        """ Flush the converted relations
        """
        # Always flush relations to be sure
        self.logger.info('Convert relations %s %s %s', etype_from, rtype, etype_to)
        self.flush_relations()
        if uri_label_from:
            self.fill_uri_eid_table(etype_from, uri_label_from)
        if uri_label_to:
            self.fill_uri_eid_table(etype_to, uri_label_to)
        if self.schema.rschema(rtype).inlined:
            self.logger.warning("Can't insert inlined relation %s", rtype)
            return
        if uri_label_from and uri_label_to:
            sql = '''INSERT INTO %(r)s_relation (eid_from, eid_to) SELECT DISTINCT O1.eid, O2.eid
            FROM %(r)s_relation_iid_tmp AS T, uri_eid_%(ef)s as O1, uri_eid_%(et)s as O2
            WHERE O1.uri=T.uri_from AND O2.uri=T.uri_to AND NOT EXISTS (
            SELECT 1 FROM %(r)s_relation AS TT WHERE TT.eid_from=O1.eid AND TT.eid_to=O2.eid);
            '''
        elif uri_label_to:
            sql = '''INSERT INTO %(r)s_relation (eid_from, eid_to) SELECT DISTINCT
            CAST(T.uri_from AS INTEGER), O1.eid
            FROM %(r)s_relation_iid_tmp AS T, uri_eid_%(et)s as O1
            WHERE O1.uri=T.uri_to AND NOT EXISTS (
            SELECT 1 FROM %(r)s_relation AS TT WHERE
            TT.eid_from=CAST(T.uri_from AS INTEGER) AND TT.eid_to=O1.eid);
            '''
        elif uri_label_from:
            sql = '''INSERT INTO %(r)s_relation (eid_from, eid_to) SELECT DISTINCT O1.eid, T.uri_to
            O1.eid, CAST(T.uri_to AS INTEGER)
            FROM %(r)s_relation_iid_tmp AS T, uri_eid_%(ef)s as O1
            WHERE O1.uri=T.uri_from AND NOT EXISTS (
            SELECT 1 FROM %(r)s_relation AS TT WHERE
            TT.eid_from=O1.eid AND TT.eid_to=CAST(T.uri_to AS INTEGER));
            '''
        try:
            self.sql(sql % {'r': rtype.lower(),
                            'et': etype_to.lower() if etype_to else u'',
                            'ef': etype_from.lower() if etype_from else u''})
        except Exception as ex:
            self.logger.error("Can't insert relation %s: %s", rtype, ex)

    # SQL utilities #########################################################

    def _drop_all_constraints(self):
        etypes_tables = ('cw_%s' % eschema.type.lower() for eschema in self.schema.entities()
                         if not eschema.final)
        rtypes_tables = ('%s_relation' % rschema.type.lower() for rschema in self.schema.relations()
                         if rschema_has_table(rschema, skip_relations=PURE_VIRTUAL_RTYPES))
        for tablename in chain(etypes_tables, rtypes_tables, ('entities',)):
            self._dbh.drop_constraints(tablename)

    def _drop_metatables_constraints(self):
        """ Drop all the constraints for the meta data"""
        for tablename in ('created_by_relation', 'owned_by_relation',
                          'is_instance_of_relation', 'is_relation',
                          'entities'):
            self._dbh.drop_indexes(tablename)

    def restart_eid_sequence(self, start_eid):
        self.sql(self._cnx.repo.system_source.dbhelper.sql_restart_numrange(
            'entities_id_seq', initial_value=start_eid))
        self._cnx.commit()

    # store api ################################################################

    def prepare_insert_entity(self, etype, **kwargs):
        """Given an entity type, attributes and inlined relations, returns the inserted entity's
        eid.
        """
        if not self.slave_mode and etype not in self._initialized:
            self._initialized.add(etype)
            self._dbh.drop_indexes('cw_%s' % etype.lower())
            self.sql('CREATE TABLE IF NOT EXISTS cwmassive_initialized'
                     '(retype text, type varchar(128))')
            self.sql("INSERT INTO cwmassive_initialized VALUES (%(e)s, 'etype')", {'e': etype})
            self.sql('ALTER TABLE cw_%s ADD COLUMN extid VARCHAR(256)' % etype.lower())
        attrs = self.metagen.base_etype_attrs(etype)
        data = copy(attrs)  # base_etype_attrs is @cached, a copy is necessary
        data.update(kwargs)
        if 'eid' not in data:
            # If eid is not given and the eids sequence is set, use the value from the sequence
            eid = self.get_next_eid()
            data['eid'] = eid
        # XXX default values could be set once for all in base entity
        default_values = self.default_values[etype]
        missing_keys = set(default_values) - set(data)
        data.update((key, default_values[key]) for key in missing_keys)
        extid = self.metagen.entity_extid(etype, data['eid'], data)
        if extid is not None:
            extid = b64encode(extid).decode('ascii')
        data['extid'] = extid
        self.metagen.init_entity_attrs(etype, data['eid'], data)
        self._data_entities[etype].append(data)
        return data['eid']

    def prepare_insert_relation(self, eid_from, rtype, eid_to, **kwargs):
        """Insert into the database a  relation ``rtype`` between entities with eids ``eid_from``
        and ``eid_to``.
        """
        assert not self._cnx.vreg.schema.rschema(rtype).inlined
        if not self.slave_mode and rtype not in self._initialized:
            self._initialized.add(rtype)
            self._dbh.drop_indexes('%s_relation' % rtype.lower())
            self.sql('CREATE TABLE %s_relation_tmp (eid_from integer, eid_to integer)'
                     % rtype.lower())
            self.sql('CREATE TABLE IF NOT EXISTS cwmassive_initialized'
                     '(retype text, type varchar(128))')
            self.sql("INSERT INTO cwmassive_initialized VALUES (%(e)s, 'rtype')", {'e': rtype})
        self._data_relations[rtype].append({'eid_from': eid_from, 'eid_to': eid_to})

    def flush(self):
        """Flush the data"""
        self.flush_entities()
        self.flush_internal_relations()
        self.flush_relations()

    def commit(self):
        """Commit the database transaction."""
        self.on_commit()
        super(MassiveObjectStore, self).commit()

    def finish(self):
        """Remove temporary tables and columns."""
        self.logger.info("Start cleaning")
        if self.slave_mode:
            raise RuntimeError('Store cleanup is not allowed in slave mode')
        self.logger.info("Start cleaning")
        # Cleanup relations tables
        for etype in self._init_uri_eid:
            self.sql('DROP TABLE uri_eid_%s' % etype.lower())
        # Remove relations tables
        for rtype in self._uri_rtypes:
            self.sql('DROP TABLE %(r)s_relation_iid_tmp' % {'r': rtype})
        # Get all the initialized etypes/rtypes
        if self._dbh.table_exists('cwmassive_initialized'):
            cu = self.sql('SELECT retype, type FROM cwmassive_initialized')
            for retype, _type in cu.fetchall():
                self.logger.info('Cleanup for %s' % retype)
                if _type == 'etype':
                    self.sql('ALTER TABLE cw_%s DROP COLUMN extid' % retype)
                elif _type == 'rtype':
                    # Cleanup relations tables
                    self._cleanup_relations(retype)
                self.sql('DELETE FROM cwmassive_initialized WHERE retype = %(e)s',
                         {'e': retype})
        self._dbh.restore_indexes_and_constraints()
        # Delete the meta data table
        self.sql('DROP TABLE IF EXISTS cwmassive_initialized')
        self.commit()

    # FLUSH #################################################################

    def on_commit(self):
        if self.on_commit_callback:
            self.on_commit_callback()

    def on_rollback(self, exc, etype, data):
        if self.on_rollback_callback:
            self.on_rollback_callback(exc, etype, data)
            self._cnx.rollback()
        else:
            raise exc

    def flush_internal_relations(self):
        """ Flush the relations data
        """
        for rtype, data in self._data_relations.items():
            if not data:
                # There is no data for these etype for this flush round.
                continue
            buf = pgstore._create_copyfrom_buffer(data, ('eid_from', 'eid_to'))
            if not buf:
                # The buffer is empty. This is probably due to error in _create_copyfrom_buffer
                raise ValueError
            cursor = self._cnx.cnxset.cu
            # Push into the tmp table
            cursor.copy_from(buf, '%s_relation_tmp' % rtype.lower(),
                             null='NULL', columns=('eid_from', 'eid_to'))
            # Clear data cache
            self._data_relations[rtype] = []

    def flush_entities(self):
        """ Flush the entities data
        """
        for etype, data in self._data_entities.items():
            if not data:
                # There is no data for these etype for this flush round.
                continue
            # XXX It may be interresting to directly infer the columns' names from the schema
            # XXX For now, the _create_copyfrom_buffer does a "row[column]"
            # which can lead to a key error.
            # Thus we should create dictionary with all the keys.
            columns = set()
            for d in data:
                columns.update(d)
            _data = []
            _base_data = dict.fromkeys(columns)
            for d in data:
                _d = _base_data.copy()
                _d.update(d)
                _data.append(_d)
            buf = pgstore._create_copyfrom_buffer(_data, columns)
            if not buf:
                # The buffer is empty. This is probably due to error in _create_copyfrom_buffer
                raise ValueError('Error in buffer creation for etype %s' % etype)
            columns = ['cw_%s' % attr if attr != 'extid' else attr
                       for attr in columns]
            cursor = self._cnx.cnxset.cu
            try:
                cursor.copy_from(buf, 'cw_%s' % etype.lower(), null='NULL', columns=columns)
            except Exception as exc:
                self.on_rollback(exc, etype, data)
            # Clear data cache
            self._data_entities[etype] = []
            if not self.slave_mode:
                self._insert_etype_metadata(etype)

    def _cleanup_relations(self, rtype):
        """ Cleanup rtype table """
        # Push into relation table while removing duplicate
        self.sql('INSERT INTO %(r)s_relation (eid_from, eid_to) SELECT DISTINCT '
                 'T.eid_from, T.eid_to FROM %(r)s_relation_tmp AS T '
                 'WHERE NOT EXISTS (SELECT 1 FROM %(r)s_relation AS TT WHERE '
                 'TT.eid_from=T.eid_from AND TT.eid_to=T.eid_to);' % {'r': rtype})
        # Drop temporary relation table
        self.sql('DROP TABLE %(r)s_relation_tmp' % {'r': rtype.lower()})

    def _insert_etype_metadata(self, etype):
        """Massive insertion of meta data for a given etype, based on SQL statements.
        """
        # insert standard metadata relations
        for rtype, eid in self.metagen.base_etype_rels(etype).items():
            self._insert_meta_relation(etype, eid, '%s_relation' % rtype)
        # insert cw_source, is and is_instance_of relations (normally handled by the system source)
        self._insert_meta_relation(etype, self.metagen.source.eid, 'cw_source_relation')
        eschema = self.schema[etype]
        self._insert_meta_relation(etype, eschema.eid, 'is_relation')
        for parent_eschema in chain(eschema.ancestors(), [eschema]):
            self._insert_meta_relation(etype, parent_eschema.eid, 'is_instance_of_relation')
        # finally insert records into the entities table
        self.sql("INSERT INTO entities (eid, type, asource, extid) "
                 "SELECT cw_eid, '%s', '%s', extid FROM cw_%s "
                 "WHERE NOT EXISTS (SELECT 1 FROM entities WHERE eid=cw_eid)"
                 % (etype, self.metagen.source.uri, etype.lower()))

    def _insert_meta_relation(self, etype, eid_to, rtype):
        self.sql("INSERT INTO %s (eid_from, eid_to) SELECT cw_eid, %s FROM cw_%s "
                 "WHERE NOT EXISTS (SELECT 1 FROM entities WHERE eid=cw_eid)"
                 % (rtype, eid_to, etype.lower()))


def get_size_constraints(schema):
    """analyzes yams ``schema`` and returns the list of size constraints.

    The returned value is a dictionary mapping entity types to a
    sub-dictionnaries mapping attribute names -> max size.
    """
    size_constraints = {}
    # iterates on all entity types
    for eschema in schema.entities():
        # for each entity type, iterates on attribute definitions
        size_constraints[eschema.type] = eschema_constraints = {}
        for rschema, aschema in eschema.attribute_definitions():
            # for each attribute, if a size constraint is found,
            # append it to the size constraint list
            maxsize = None
            rdef = rschema.rdef(eschema, aschema)
            for constraint in rdef.constraints:
                if isinstance(constraint, SizeConstraint):
                    maxsize = constraint.max
                    eschema_constraints[rschema.type] = maxsize
    return size_constraints


def get_default_values(schema):
    """analyzes yams ``schema`` and returns the list of default values.

    The returned value is a dictionary mapping entity types to a
    sub-dictionnaries mapping attribute names -> default values.
    """
    default_values = {}
    # iterates on all entity types
    for eschema in schema.entities():
        # for each entity type, iterates on attribute definitions
        default_values[eschema.type] = eschema_constraints = {}
        for rschema, _ in eschema.attribute_definitions():
            # for each attribute, if a size constraint is found,
            # append it to the size constraint list
            if eschema.default(rschema.type) is not None:
                eschema_constraints[rschema.type] = eschema.default(rschema.type)
    return default_values


class PGHelper(object):
    """This class provides some helper methods to manipulate a postgres database metadata (index and
    constraints).
    """

    def __init__(self, cnx):
        self.sql = cnx.system_sql
        # Deals with pg schema, see #3216686
        pg_schema = cnx.repo.config.system_source_config.get('db-namespace') or 'public'
        self.pg_schema = pg_schema

    def drop_indexes(self, tablename):
        """Drop indexes and constraints, storing them in a table for later restore."""
        # Create a table to save the constraints, it allows reloading even after crash
        self.sql('CREATE TABLE IF NOT EXISTS cwmassive_constraints(sql TEXT, insert_order SERIAL)')
        indexes = self.table_indexes(tablename)
        for name, query in indexes.items():
            self.sql('INSERT INTO cwmassive_constraints(sql) VALUES (%(sql)s)', {'sql': query})
            self.sql('DROP INDEX %s' % name)

    def drop_constraints(self, tablename):
        self.sql('CREATE TABLE IF NOT EXISTS cwmassive_constraints(sql TEXT, insert_order SERIAL)')
        constraints = self.table_constraints(tablename)
        for name, query in constraints.items():
            self.sql('INSERT INTO cwmassive_constraints(sql) VALUES (%(sql)s)', {'sql': query})
            self.sql('ALTER TABLE %s DROP CONSTRAINT %s' % (tablename, name))

    def restore_indexes_and_constraints(self):
        """Restore indexes and constraints."""
        if not self.table_exists('cwmassive_constraints'):
            self.logger.info('The table cwmassive_constraints does not exist')
            return
        cu = self.sql('SELECT sql, insert_order FROM cwmassive_constraints '
                      'ORDER BY insert_order DESC')
        for query, order in cu.fetchall():
            self.sql(query)
            self.sql('DELETE FROM cwmassive_constraints WHERE insert_order=%(order)s',
                     {'order': order})
        self.sql('DROP TABLE cwmassive_constraints')

    def table_exists(self, tablename):
        """Return True if the given table already exists in the database."""
        cu = self.sql('SELECT 1 from information_schema.tables '
                      'WHERE table_name=%(t)s AND table_schema=%(s)s',
                      {'t': tablename, 's': self.pg_schema})
        return bool(cu.fetchone())

    def table_indexes_constraints(self, tablename):
        """Return one dictionary with all indexes by name, another with all constraints by name,
        for the given table.
        """
        indexes = self.table_indexes(tablename)
        constraints = self.table_constraints(tablename)
        _indexes = {}
        for name, query in indexes.items():
            # Remove pkey indexes (automatically created by constraints)
            # Specific cases of primary key, see #3224079
            if name not in constraints:
                _indexes[name] = query
        return _indexes, constraints

    def table_indexes(self, tablename):
        """Return a dictionary of indexes {index name: index sql}, constraints included."""
        indexes = {}
        for name in self._index_names(tablename):
            indexes[name] = self._index_sql(name)
        return indexes

    def table_constraints(self, tablename):
        """Return a dictionary of constraints {constraint name: constraint sql}."""
        constraints = {}
        for name in self._constraint_names(tablename):
            query = self._constraint_sql(name)
            constraints[name] = 'ALTER TABLE %s ADD CONSTRAINT %s %s' % (tablename, name, query)
        return constraints

    def _index_names(self, tablename):
        """Return the names of all indexes in the given table (including constraints.)"""
        cu = self.sql("SELECT c.relname FROM pg_catalog.pg_class c "
                      "JOIN pg_catalog.pg_index i ON i.indexrelid = c.oid "
                      "JOIN pg_catalog.pg_class c2 ON i.indrelid = c2.oid "
                      "LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner "
                      "LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace "
                      "WHERE c.relkind IN ('i','') "
                      " AND c2.relname = %(t)s "
                      " AND i.indisprimary = FALSE "
                      " AND n.nspname NOT IN ('pg_catalog', 'pg_toast') "
                      " AND pg_catalog.pg_table_is_visible(c.oid);", {'t': tablename})
        return [name for name, in cu.fetchall()]

    def _constraint_names(self, tablename):
        """Return the names of all constraints in the given table."""
        cu = self.sql("SELECT i.conname FROM pg_catalog.pg_class c "
                      "JOIN pg_catalog.pg_constraint i ON i.conrelid = c.oid "
                      "JOIN pg_catalog.pg_class c2 ON i.conrelid=c2.oid "
                      "LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner "
                      "LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace "
                      "WHERE c2.relname = %(t)s "
                      "AND n.nspname NOT IN ('pg_catalog', 'pg_toast') "
                      "AND pg_catalog.pg_table_is_visible(c.oid)", {'t': tablename})
        return [name for name, in cu.fetchall()]

    def _index_sql(self, name):
        """Return the SQL to be used to recreate the index of the given name."""
        return self.sql('SELECT pg_get_indexdef(c.oid) FROM pg_catalog.pg_class c '
                        'LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace '
                        'WHERE c.relname = %(r)s AND n.nspname=%(n)s',
                        {'r': name, 'n': self.pg_schema}).fetchone()[0]

    def _constraint_sql(self, name):
        """Return the SQL to be used to recreate the constraint."""
        return self.sql('SELECT pg_get_constraintdef(c.oid) FROM pg_catalog.pg_constraint c '
                        'LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.connamespace '
                        'WHERE c.conname = %(r)s AND n.nspname=%(n)s',
                        {'r': name, 'n': self.pg_schema}).fetchone()[0]



add_entity_type('AgentRelation')

drop_relation_type('acted_on_behalf_of')

sync_schema_props_perms('Agent')

rename_attribute('Agent', 'identifier', 'isni')

add_entity_type('AgentPlace')

for a_eid, p_aeid in rql('Any X,A WHERE X postal_address A'):
    create_entity('AgentPlace', place_agent=a_eid, place_address=p_eid)
commit()

for agentkind in find_entities('AgentKind'):
    agentkind.cw_set(cwuri='agentkind/' + agentkind.name)
commit()


add_entity_type('AgentFunction')
for aeid, function in rql('Any X,F WHERE EXISTS(X function F), X is Agent').rows:
    create_entity('AgentFunction', name=function, function_agent=aeid)
commit()

add_entity_type('EACResourceRelation')

add_relation_type('equivalent_concept')

add_entity_type('EACSource')



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from cubicweb.devtools.testlib import CubicWebTC


class ArkGeneratorTC(CubicWebTC):
    def test_eid(self):
        with self.admin_access.repo_cnx() as cnx:
            generator = self.vreg['adapters'].select('IARKGenerator', cnx, eid=12)
            self.assertEqual(generator.generate_ark(), 'saemref-test/000000012')

    def test_ext_identifier(self):
        with self.admin_access.repo_cnx() as cnx:
            generator = self.vreg['adapters'].select('IARKGenerator', cnx)
            self.assertEqual(generator.generate_ark(), 'saemref-test/ext-000000001')


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



sync_schema_props_perms('place_address')

add_relation_definition('Agent', 'agent_user', 'CWUser')

kind = cnx.find('AgentKind', name=u'person').one()
for user in find('CWUser').entities():
    if user.login == 'anon':
        continue
    create_entity('Agent', name=user.login, agent_user=user, agent_kind=kind)
commit()

# specialize generic EACInformation
add_entity_type('Mandate')
add_entity_type('LegalStatus')
add_entity_type('Structure')
add_entity_type('History')

for eacinfo in find_entities('EACInformation'):
    if eacinfo.type == u'mandate':
        create_entity('Mandate', term=eacinfo.term,
                      description=eacinfo.description,
                      description_format=eacinfo.description_format,
                      start_date=eacinfo.start_date,
                      end_date=eacinfo.end_date,
                      mandate_agent=eacinfo.eac_information_agent)
    elif eacinfo.type == u'legalStatus':
        create_entity('LegalStatus', term=eacinfo.term,
                      description=eacinfo.description,
                      description_format=eacinfo.description_format,
                      start_date=eacinfo.start_date,
                      end_date=eacinfo.end_date,
                      legal_status_agent=eacinfo.eac_information_agent)
    elif eacinfo.type == u'biogHist':
        create_entity('History', text=eacinfo.description,
                      text_format=eacinfo.description_format,
                      history_agent=eacinfo.eac_information_agent)
    elif eacinfo.type == u'structureOrGenealogy':
        create_entity('Structure', description=eacinfo.description,
                      description_format=eacinfo.description_format,
                      structure_agent=eacinfo.eac_information_agent)

drop_entity_type('EACInformation')
drop_relation_type('eac_information_agent')


# Description changed
sync_schema_props_perms('agent_relationship_from')
sync_schema_props_perms('agent_relationship_to')

sync_schema_props_perms('ExternalUri', syncprops=False)

sync_schema_props_perms('start_date')
sync_schema_props_perms('end_date')


add_cube('prov')
sync_schema_props_perms('Agent')

eac2prov = {"created": u'create',
            "derived": u'create',
            "revised": u'replace',
            "updated": u'modify',
           }

prov_activity_types = schema['Activity'].rdef('type').constraint_by_type('StaticVocabularyConstraint').values

for event in find_entities('MaintenanceEvent'):
    event_type = event.event_type
    if event_type not in prov_activity_types:
        if event_type in eac2prov:
            event_type = eac2prov[event_type]
        else:
            print ('could not find a correspondance for "%s" maintenance event '
                   'in PROV-O terminology, %s will be dropped.' % (event_type, event))
            continue
    assert event_type in prov_activity_types, event_type
    create_entity('Activity', type=event_type,
                  description=event.description,
                  description_format=event.description_format,
                  start=event.datetime,
                  end=event.datetime,
                  associated_with=event.event_agent,
                  generated=event.event_concerns,
                  used=event.event_concerns,
                 )
    commit()

drop_entity_type('MaintenanceEvent')


add_attribute('MirrorEntity', 'error_flag')


for rtype in ('function_agent', 'phone_number', 'place_address', 'place_agent',
              'mandate_agent', 'structure_agent', 'history_agent',
              'resource_relation_agent', 'source_agent'):
    sync_schema_props_perms(rtype, syncperms=False)


add_cube('relationwidget')

# cubicweb-skos had no release before, adding migration steps here instead.
add_relation_type('exact_match')
add_relation_type('close_match')
sync_schema_props_perms()  # too many changes in skos (fti, constraints, etc.)
